#!/usr/bin/python3
import sys
from pathlib import Path
from typing import Tuple
from PyQt5 import QtCore, QtWidgets, QtGui

try:
    from ..backend import Tuples, configEnum, logger
    from ..plugins import antidote
except (ValueError, ImportError):
    from backend import Tuples, configEnum, logger
    from plugins import antidote

log = logger.new(__name__)


class EncryptionMenu(QtWidgets.QMenu):
    def __init__(self, note):
        super().__init__(note)
        self.setTitle("Encryption")
        self.setIcon(note.icons.menu["key_encrypt"])
        enabled = bool(note.content) and note.hasClearText()
        self.setEnabled(enabled)

        actions = (
            (note.icons.menu["key_encrypt"], "Encrypt with master key", note.keyring.setMaster),
            (note.icons.menu["key_encrypt"], "Encrypt with unique key", note.keyring.setUnique),
            (note.icons.menu["key_replace"], "Replace unique key", note.keyring.replace),
            (note.icons.menu["key_decrypt"], "Remove encryption", note.keyring.remove),
            (note.icons.menu["lock"], "Logout", note.keyring.logout),
        )
        for action in actions:
            icon, text, func = action
            action = QtWidgets.QAction(icon, text, self)
            action.triggered.connect(func)
            self.addAction(action)


class ModeMenu(QtWidgets.QMenu):
    def __init__(self, note):
        super().__init__(note)
        self.core = note.core
        self.note = note

        action = note.actions["mode"]
        self.setTitle("Mode")
        self.setIcon(action.icon.menu)

        for mode in ("plain", "html"):
            if mode != note.mode:
                self.addAction(self.ModeAction(self.note, mode))

        if "QTermWidget" in sys.modules:
            if note.mode != "vim":
                self.addAction(self.ModeAction(self.note, "vim"))
            if note.mode != "console":
                self.addAction(self.ModeAction(self.note, "console"))

        self.addSeparator()
        self.addAction(self.ModeActionSetDefault(self.note))

    class ModeAction(QtWidgets.QAction):
        def __init__(self, note, mode: str):
            super().__init__(note)
            self.note = note
            self.core = note.core
            self.mode = mode
            self.triggered.connect(self._triggered)
            self.modes = {
                "console": Tuples.Mode(note.icons.menu["console"], "Console"),
                "html": Tuples.Mode(note.icons.menu["html_regular"], "Rich text"),
                "plain": Tuples.Mode(note.icons.menu["plain_regular"], "Plain text"),
                "vim": Tuples.Mode(note.icons.menu["vim"], "Vim"),
            }
            self.setIcon(self.modes[mode].icon)
            self.setText(self.modes[mode].label)

        def _triggered(self):
            """Handler for left click event, sets editing mode"""
            self.core.notes.mode(self.note.path, self.mode)

    class ModeActionSetDefault(QtWidgets.QAction):
        def __init__(self, note):
            super().__init__(note)
            self.note = note
            self.core = note.core
            self.setText("Set as default")
            self.setIcon(note.icons.menu["add"])
            self.triggered.connect(self._triggered)

        def _triggered(self):
            """Sets default mode for new notes"""
            self.core.sdb["profile default"]["mode"] = self.note.mode
            log.info(f"Default mode set to '{self.note.mode}'")


class CSSMenu(QtWidgets.QMenu):
    def __init__(self, note):
        super().__init__(note)
        self.note = note
        self.core = note.core
        self.aboutToShow.connect(self.refresh)

    def refresh(self):
        """Finds the installed CSS files and updates the menu entries"""
        self.clear()
        files = configEnum(self.type, ignored=self.core.sdb["general"]["ignored css"])
        for f in sorted(files, key=lambda item: item.stem.lower()):
            self.addAction(self.action(self.note, path=f))

        self.addSeparator()
        self.addAction(self.default(self.note))


class CSSAction(QtWidgets.QAction):
    def __init__(self, note, path: Path):
        super().__init__(note)
        self.note = note
        self.core = note.core
        self.path = path
        self.setText(path.stem.capitalize())
        self.triggered.connect(self._triggered)

    def _save(self):
        """Save newly selected style to profile"""
        self.core.pdb[self.note.id][self.type] = self.path.name

    def _triggered(self):
        """Handler for left click event"""
        self._save()
        self.note.decorate()


class CSSActionSetDefault(QtWidgets.QAction):
    def __init__(self, note):
        super().__init__(note)
        self.note = note
        self.core = note.core
        self.setText("Set as default")
        self.setIcon(note.icons.menu["add"])
        self.triggered.connect(self._triggered)

    def _triggered(self):
        """Handler for left click event, sets style to default"""
        css = self.core.pdb[self.note.id][self.type]
        self.core.sdb["profile default"][self.type] = css
        log.info(f"Default {self.type} set to '{css}'")


class ExportDialog(QtWidgets.QFileDialog):
    """Export browser dialog (all modes)"""

    def __init__(self, core, path: Path, nameFilters: list):
        super().__init__()
        lastPath = core.sdb["runtime"]["export directory"]
        self.setDirectory(lastPath)
        self.setWindowTitle(f"Export {path.stem} as ...")
        self.setWindowFlags(QtCore.Qt.WindowStaysOnTopHint)
        self.setOptions(self.DontUseNativeDialog)
        self.setAcceptMode(self.AcceptSave)
        self.setFileMode(self.AnyFile)
        self.setViewMode(self.Detail)
        self.setNameFilters(nameFilters)
        self.selectFile(path.with_suffix("").name)
        self.setModal(True)
        for listView in self.findChildren(QtWidgets.QListView):
            listView.hide()


class PaletteMenu(CSSMenu):
    def __init__(self, note):
        super().__init__(note)
        self.note = note
        self.setTitle("Palette")
        self.setIcon(note.actions["palette"].icon.menu)

        self.type = "palette"
        self.action = self.PaletteAction
        self.default = self.PaletteActionSetDefault

    class PaletteAction(CSSAction):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.type = "palette"
            try:
                self._setIcon()
            except AttributeError:  # Failed to extract fg/bg from stylesheet
                self.setIcon(self.core.icons["rename"])

        def _colors(self, path: Path) -> Tuple[str, str]:
            """Fetches the colors of a preview icon"""
            css = self.core.getNoteDecorationsCSS(path)
            fg, bg = css["preview-fg"], css["preview-bg"]
            return fg, bg

        def _setIcon(self):
            """Apply custom colors for the preview icons"""
            fg, bg = self._colors(self.path)
            pixmap = self.core.icons["rename"].pixmap(16, 16)
            painter = QtGui.QPainter(pixmap)
            painter.setCompositionMode(painter.CompositionMode_Xor)
            painter.fillRect(pixmap.rect(), QtGui.QColor(bg))
            painter.setCompositionMode(painter.CompositionMode_DestinationOver)
            painter.fillRect(pixmap.rect(), QtGui.QColor(fg))
            painter.end()
            self.setIcon(QtGui.QIcon(pixmap))

    class PaletteActionSetDefault(CSSActionSetDefault):
        def __init__(self, note):
            super().__init__(note)
            self.type = "palette"


class StyleMenu(CSSMenu):
    def __init__(self, note):
        super().__init__(note)
        self.note = note
        self.setTitle("Style")
        self.setIcon(note.actions["style"].icon.menu)

        self.type = "style"
        self.action = self.StyleAction
        self.default = self.StyleActionSetDefault

    class StyleAction(CSSAction):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.type = "style"
            self.setIcon(self.note.icons.menu["aspect_ratio"])

    class StyleActionSetDefault(CSSActionSetDefault):
        def __init__(self, note):
            super().__init__(note)
            self.type = "style"


class ToolMenu(QtWidgets.QMenu):
    def __init__(self, note):
        super().__init__()
        self.note = note
        self.core = note.core
        self.aboutToShow.connect(self._refresh)
        self.items = {
            "separator": self.addSeparator,
            "mode": ModeMenu,
            "palette": PaletteMenu,
            "style": StyleMenu,
        }
        if note.mode in ("plain", "html"):
            self.items["encryption"] = EncryptionMenu

    def _refresh(self):
        """Updates tools menu on request"""
        self.clear()
        menu = self.core.sdb["context menus"][self.note.mode]
        for key in menu:
            try:  # Add submenus
                menu = self.items[key](self.note)
                self.addMenu(menu)
            except KeyError:  # Add actions
                try:
                    if key.startswith("antidote") and not antidote.isInstalled():
                        continue
                    action = self.note.actions[key]
                    tool = self.ToolAction(self.note, action)
                    self.addAction(tool)
                except KeyError:
                    log.exception(f"Invalid action '{key}'")
            except TypeError:
                self.addSeparator()

    class ToolAction(QtWidgets.QAction):
        def __init__(self, note, action: Tuples.Action):
            super().__init__(note)
            self.note = note
            self.action = action
            self.setText(action.label)
            self.setIcon(self.action.icon.menu)
            self.triggered.connect(self._triggered)

        def _triggered(self, event):
            """Handler for left click event, calls an action"""
            log.info(f"{self.note.id} : {self.action.label}")
            try:
                self.action.call(self.note.path)
            except TypeError:
                self.action.call()



#!/usr/bin/python3
from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtWidgets import QLabel, QSizePolicy, QSpacerItem

try:
    from ..backend import logger
    from ..notes import NoteSetup
    from ..settings.base import Page, MODE_TO_KEY, KEY_TO_MODE
except (ValueError, ImportError):
    from backend import logger
    from notes import NoteSetup
    from settings.base import Page, MODE_TO_KEY, KEY_TO_MODE

log = logger.new(__name__)


class PageMenu(Page):
    def __init__(self, core, db: dict):
        super().__init__(core, db)
        sizePolicy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        sizePolicy.setVerticalStretch(1)
        self.availableList = self.AvailablesList()
        self.availableList.setSizePolicy(sizePolicy)
        self.availableList.setIconSize(QtCore.QSize(16, 16))
        self.availableList.setToolTip("Drag an item from the available to the selected elements")

        self.selectedList = self.SelectedList()
        self.selectedList.setIconSize(QtCore.QSize(16, 16))
        self.selectedList.dropped.connect(self._save)
        self.selectedList.setToolTip(
            "Drag and drop an item inside of the selected elements\n"
            "to change its position, drop it outside to delete it"
        )

        self.setAcceptDrops(True)  # Act as a drop dump to discard item from selected list
        self._layout()

    def dragEnterEvent(self, event: QtGui.QDragEnterEvent):
        event.setDropAction(QtCore.Qt.MoveAction)
        event.accept()

    def dropEvent(self, event: QtGui.QDropEvent):
        """Accepts drops so the whole widget act as a drop dump"""
        if event.source() is self.selectedList:
            event.setDropAction(QtCore.Qt.MoveAction)
            event.accept()
            QtCore.QTimer.singleShot(0, self._save)  # Wait for the item to be deleted

    def populate(self):
        """Parses availables actions/database keys to generate QListWidget items"""
        self.availableList.clear()
        for key, action in sorted(self.actions.items()):
            item = self._item(key, action)
            self.availableList.addItem(item)

        self.selectedList.clear()
        for key in self.menus[self.index]:
            try:
                action = self.actions[key]
                item = self._item(key, action)
                self.selectedList.addItem(item)
            except KeyError:
                log.exception(f"Invalid action '{key}'")

    def _item(self, key: str, action: object) -> QtWidgets.QListWidgetItem:
        """Converts a key/action pair to a QListWidgetItem"""
        item = QtWidgets.QListWidgetItem()
        item.setText(action.label)
        item.setIcon(action.icon)
        item.setData(QtCore.Qt.UserRole, key)
        return item

    def _layout(self):
        """Setups the page layout"""
        frame = QtWidgets.QWidget()
        sizePolicy = QSizePolicy(QSizePolicy.Maximum, QSizePolicy.Preferred)
        label = QtWidgets.QLabel("Current menu :  ", frame)
        label.setSizePolicy(sizePolicy)
        self.menuBox = QtWidgets.QComboBox(frame)
        self.menuBox.setToolTip("Select a menu to edit")

        menuLayout = QtWidgets.QHBoxLayout(frame)
        menuLayout.setContentsMargins(0, 0, 0, 0)
        menuLayout.addWidget(label)
        menuLayout.addWidget(self.menuBox)

        vspacer = QSpacerItem(0, 15, QSizePolicy.Minimum, QSizePolicy.Fixed)
        hspacer = QSpacerItem(5, 0, QSizePolicy.Fixed, QSizePolicy.Minimum)
        layout = QtWidgets.QGridLayout(self)
        layout.addWidget(frame, 0, 0, 1, 3)
        layout.addWidget(QLabel("Available elements"), 2, 0, 1, 1)
        layout.addWidget(QLabel("Selected elements"), 2, 2, 1, 1)
        layout.addWidget(self.availableList, 3, 0, 2, 1)
        layout.addItem(vspacer, 1, 0, 1, 1)
        layout.addWidget(self.selectedList, 3, 2, 2, 1)
        layout.addItem(hspacer, 3, 1, 1, 1)

    def _save(self):
        """Save selected list widget content to settings database"""
        selectedList = self.selectedList
        itemsCount = selectedList.count()
        items = [selectedList.item(i).data(QtCore.Qt.UserRole) for i in range(itemsCount)]
        self.menus[self.index] = items

    class AvailablesList(QtWidgets.QListWidget):
        def __init__(self):
            super().__init__()
            self.setDragDropMode(QtWidgets.QAbstractItemView.DragOnly)

    class SelectedList(QtWidgets.QListWidget):
        dropped = QtCore.pyqtSignal()

        def __init__(self):
            super().__init__()
            self.setDragDropMode(QtWidgets.QAbstractItemView.DragDrop)
            self.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)

        def dropEvent(self, event: QtGui.QDropEvent):
            """Accepts dropped items"""
            event.setDropAction(QtCore.Qt.MoveAction)
            super().dropEvent(event)
            self.dropped.emit()


class PageCoreMenu(PageMenu):
    def __init__(self, *args):
        super().__init__(*args)
        self.menuBox.addItems(["Tray", "Browser"])
        self.menuBox.currentTextChanged.connect(self.setMenu)
        self.setMenu("tray")

    def setMenu(self, menu: str):
        """Updates QListWidget content as the QComboBox content is changed"""
        menu = menu.lower()
        self.menus, self.index = self.db["core menus"], menu
        self.actions = self.core.actions.tray if menu == "tray" else self.core.actions.browser
        self.populate()


class PageNoteMenu(PageMenu):
    def __init__(self, *args):
        super().__init__(*args)
        current = self.db["profile default"]["mode"]
        current = KEY_TO_MODE[current]
        self.menuBox.addItems([menu for menu in MODE_TO_KEY])
        self.menuBox.setCurrentText(current)
        self.menuBox.currentTextChanged.connect(self.setMenu)
        self.setMenu(current)

    def setMenu(self, menu: str):
        """Updates QListWidget content as the QComboBox content is changed"""
        mode = MODE_TO_KEY[menu]
        self.menus, self.index = self.db[self.key], mode
        self.actions = NoteSetup.actionsFromMode(self.core, mode)
        self.populate()


class PageNoteContextMenu(PageNoteMenu):
    def __init__(self, *args):
        self.key = "context menus"
        super().__init__(*args)


class PageNoteToolbarMenu(PageNoteMenu):
    def __init__(self, *args):
        self.key = "toolbar menus"
        super().__init__(*args)



#!/usr/bin/python3
from typing import Callable, Union
from PyQt5 import QtCore, QtDBus
from PyQt5.QtGui import QTextCursor
from PyQt5.QtWidgets import QPlainTextEdit

try:
    from ...backend import logger
except (ValueError, ImportError):
    from backend import logger

log = logger.new(__name__)


class Adapter(QtDBus.QDBusAbstractAdaptor):
    QtCore.Q_CLASSINFO("D-Bus Interface", "com.druide.antidote.dbus.api")
    QtCore.Q_CLASSINFO(
        "D-Bus Introspection",
        ""
        '  <interface name="com.druide.antidote.dbus.api">\n'
        '    <method name="ActiveDocument"/>\n'
        '    <method name="CorrigeDansTexteur">\n'
        '      <arg direction="in" type="i" name="leDebut"/>\n'
        '      <arg direction="in" type="i" name="laFin"/>\n'
        '      <arg direction="in" type="s" name="laChaine"/>\n'
        '      <arg direction="in" type="b" name="automatique"/>\n'
        "    </method>\n"
        '    <method name="DonneBloc">\n'
        '      <arg direction="in" type="i" name="leDebut"/>\n'
        '      <arg direction="in" type="i" name="laFin"/>\n'
        '      <arg direction="out" type="s" name="_ret"/>\n'
        "    </method>\n"
        '    <method name="DonneDebutSelection">\n'
        '      <arg direction="out" type="i" name="_ret"/>\n'
        "    </method>\n"
        '    <method name="DonneFinDocument">\n'
        '      <arg direction="out" type="i" name="_ret"/>\n'
        "    </method>\n"
        '    <method name="DonneFinSelection">\n'
        '      <arg direction="out" type="i" name="_ret"/>\n'
        "    </method>\n"
        '    <method name="DonnePositionFinBoite">\n'
        '      <arg direction="in" type="i" name="unePos"/>\n'
        '      <arg direction="out" type="i" name="_ret"/>\n'
        "    </method>\n"
        '    <method name="DonneRetourDeCharriot">\n'
        '      <arg direction="out" type="s" name="_ret"/>\n'
        "    </method>\n"
        '    <method name="DonneTitreDocument">\n'
        '      <arg direction="out" type="s" name="_ret"/>\n'
        "    </method>\n"
        '    <method name="PermetsRetourDeCharriot">\n'
        '      <arg direction="out" type="b" name="_ret"/>\n'
        "    </method>\n"
        '    <method name="RetourneAuTexteur"/>\n'
        '    <method name="RompsLienCorrecteur"/>\n'
        '    <method name="RompsLienTexteur"/>\n'
        '    <method name="SelectionneIntervalle">\n'
        '      <arg direction="in" type="i" name="leDebut"/>\n'
        '      <arg direction="in" type="i" name="laFin"/>\n'
        "    </method>\n"
        "  </interface>\n",
    )

    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent
        self.setAutoRelaySignals(True)

    @QtCore.pyqtSlot()
    def ActiveDocument(self):
        self.parent.activeDocument()

    @QtCore.pyqtSlot(int, int, str, bool)
    def CorrigeDansTexteur(self, debut: int, fin: int, chaine: str, auto: bool):
        self.parent.corrigeDansTexteur(debut, fin, chaine, auto)

    @QtCore.pyqtSlot(int, int, result=str)
    def DonneBloc(self, debut: int, fin: int) -> str:
        return self.parent.donneBloc(debut, fin)

    @QtCore.pyqtSlot(result=int)
    def DonneDebutSelection(self) -> int:
        return self.parent.donneDebutSelection()

    @QtCore.pyqtSlot(result=int)
    def DonneFinDocument(self) -> int:
        return self.parent.donneFinDocument()

    @QtCore.pyqtSlot(result=int)
    def DonneFinSelection(self) -> int:
        return self.parent.donneFinSelection()

    @QtCore.pyqtSlot(int, result=int)
    def DonnePositionFinBoite(self, unePos: int) -> int:
        return self.parent.donnePositionFinBoite()

    @QtCore.pyqtSlot(result=str)
    def DonneRetourDeCharriot(self) -> str:
        return self.parent.donneRetourDeCharriot()

    @QtCore.pyqtSlot(result=str)
    def DonneTitreDocument(self) -> str:
        return self.parent.donneTitreDocument()

    @QtCore.pyqtSlot(result=bool)
    def PermetsRetourDeCharriot(self) -> bool:
        return self.parent.permetsRetourDeCharriot()

    @QtCore.pyqtSlot()
    def RetourneAuTexteur(self):
        self.parent.retourneAuTexteur()

    @QtCore.pyqtSlot()
    def RompsLienCorrecteur(self):
        self.parent.rompsLienCorrecteur()

    @QtCore.pyqtSlot()
    def RompsLienTexteur(self):
        self.parent.rompsLienTexteur()

    @QtCore.pyqtSlot(int, int)
    def SelectionneIntervalle(self, debut: int, fin: int):
        self.parent.selectionneIntervalle(debut, fin)


class API:  # API Antidote 10 (2020)
    def ignoreRuntimeError(default: Union[str, int] = None) -> Callable:
        """Logs API calls and ignores RuntimeError during execution. Return a default value on failure"""

        def rewrap(func: Callable, *args) -> Callable:
            def wrap(self, *args) -> any:
                try:
                    _return = func(self, *args)
                    log.debug(f"{func.__qualname__}{args} -> {repr(_return)}")
                    return _return
                except RuntimeError:
                    log.debug(f"RuntimeError in {func.__qualname__}{args}, returned {repr(default)}")
                    return default

            return wrap

        return rewrap

    @ignoreRuntimeError()
    def activeDocument(self):
        self.body.window().raise_()

    @ignoreRuntimeError()
    def corrigeDansTexteur(self, debut: int, fin: int, chaine: str, auto: bool):
        if debut >= 0 and fin >= 0:
            textCursor = self.body.textCursor()
            textCursor.setPosition(debut)
            textCursor.movePosition(QTextCursor.NextCharacter, QTextCursor.KeepAnchor, fin - debut)
            textCursor.insertText(chaine)

        self.body.setUpdatesEnabled(not auto)
        if not auto:
            self.body.ensureCursorVisible()

    @ignoreRuntimeError(default="")
    def donneBloc(self, debut: int, fin: int) -> str:
        textCursor = self.body.textCursor()
        textCursor.movePosition(QTextCursor.Start)
        textCursor.movePosition(QTextCursor.NextCharacter, QTextCursor.MoveAnchor, debut)
        textCursor.movePosition(QTextCursor.NextCharacter, QTextCursor.KeepAnchor, fin - debut)
        return textCursor.selectedText()

    @ignoreRuntimeError(default=0)
    def donneDebutSelection(self) -> int:
        ancre = self.body.textCursor().anchor()
        position = self.body.textCursor().position()
        return min(ancre, position)

    @ignoreRuntimeError(default=0)
    def donneFinDocument(self) -> int:
        textCursor = self.body.textCursor()
        textCursor.movePosition(QTextCursor.End)
        return textCursor.position()

    @ignoreRuntimeError(default=0)
    def donneFinSelection(self) -> int:
        ancre = self.body.textCursor().anchor()
        position = self.body.textCursor().position()
        return max(ancre, position)

    def donnePositionFinBoite(self) -> int:
        return 0

    def donneRetourDeCharriot(self) -> str:
        return "\n"

    @ignoreRuntimeError(default="")
    def donneTitreDocument(self) -> str:
        return self.body.window().windowTitle()

    def permetsRetourDeCharriot(self) -> bool:
        return True

    @ignoreRuntimeError
    def retourneAuTexteur(self):
        self.body.window().activateWindow()
        self.body.window().raise_()

    @ignoreRuntimeError
    def rompsLienCorrecteur(self):
        self.body.setUpdatesEnabled(True)

    def rompsLienTexteur(self):
        pass

    @ignoreRuntimeError
    def selectionneIntervalle(self, debut: int, fin: int):
        textCursor = self.body.textCursor()
        textCursor.setPosition(debut)
        textCursor.movePosition(QTextCursor.NextCharacter, QTextCursor.KeepAnchor, fin - debut)
        self.body.setTextCursor(textCursor)


class Antidote(QtCore.QObject, API):
    ready = QtCore.pyqtSignal()

    def __init__(self, uid: str, path: str, body: QPlainTextEdit):
        super().__init__()
        self.uid = f"{uid.lower()}_{id(self)}"
        self.path = path
        self.body = body

    def init(self):
        """Initializes DBus interface"""
        self.bus = QtDBus.QDBusConnection.sessionBus()
        if self._register():
            self.__dbusAdaptor = Adapter(self)
            self.ready.emit()

    def corrector(self):
        """Launches Antidote corrector"""
        self._tool("C")

    def dictionnary(self, tool: str = "D_Defaut"):
        """Launches Antidote dictionnary"""
        options = (
            "D",
            "D_Defnitions",
            "D_Synonymes",
            "D_Antonymes",
            "D_Cooccurrences",
            "D_ChampLexical",
            "D_Conjugaison",
            "D_Famille",
            "D_Citations",
            "D_Historique",
            "D_Illustrations",
            "D_Wikipedia",
        )
        tool = tool if tool in options else "D_Defaut"
        self._tool(tool)

    def guide(self, tool: str = "G_Defaut"):
        """Launches Antidote guide"""
        options = (
            "G",
            "G_Orthographe",
            "G_Lexique",
            "G_Grammaire",
            "G_Syntaxe",
            "G_Ponctuation",
            "G_Style",
            "G_Redaction",
            "G_Typographie",
            "G_Phonetique",
            "G_Historique",
            "G_PointsDeLangue",
        )
        tool = tool if tool in options else "G_Defaut"
        self._tool(tool)

    def _tool(self, tool: str):
        p = QtCore.QProcess()
        p.setStandardOutputFile(QtCore.QProcess.nullDevice())
        p.setStandardErrorFile(QtCore.QProcess.nullDevice())
        p.setProgram(self.path)
        p.setArguments(["--module", "API_DBus:v1.0", "--identificateur", self.uid, "--outil", tool])
        p.startDetached()

    def _register(self) -> bool:
        """Registers a session DBus service"""
        obj = f"/com/druide/antidote/dbus/api/v1/{self.uid}"
        if not self.bus.registerObject(obj, self):
            log.error(f"Could not register DBus object '{obj}'")
            return False

        service = f"com.druide.antidote.dbus.api.v1.{self.uid}"
        if not self.bus.registerService(service):
            log.error(f"Could not register DBus service '{service}'")
            return False
        return True



#!/usr/bin/python3
import json
from copy import deepcopy
from pathlib import Path, PurePath
from typing import Tuple
from PyQt5 import QtCore, QtWidgets

try:
    from ..backend import configEnum, logger
    from ..backend.observer import Observer
except (ValueError, ImportError):
    from backend import configEnum, logger
    from backend.observer import Observer

log = logger.new(__name__)
REPAIR_EXCLUSIONS = ("key events",)


class AbstractDatabaseInterface:
    """Provides load and save methods to interface a dict key 'path'
    Converts a tuple to a nested dict key pointer; (k1, k2) -> dict[k1][k2]"""

    def __init__(self, db: dict, slices: Tuple):
        self.db = db
        self.slices = slices

    def _load(self) -> any:
        """Loads a database value"""
        value = self.db
        for s in self.slices:
            value = value.get(s)
        return value

    def _save(self, state: int):
        """Saves new value to database"""
        _db = self.db
        for key in self.slices[:-1]:
            _db = _db[key]
        _db[self.slices[-1]] = self.value()


class AutoCheckBox(AbstractDatabaseInterface, QtWidgets.QCheckBox):
    """Self loading/saving QCheckBox widget"""

    def __init__(self, text: str, db: dict, slices: Tuple):
        super(QtWidgets.QCheckBox, self).__init__(text)
        AbstractDatabaseInterface.__init__(self, db, slices)
        self.load()
        self.stateChanged.connect(self._save)

    def load(self):
        self.setChecked(self._load())

    def value(self) -> bool:
        return self.isChecked()


class AutoComboBox(AbstractDatabaseInterface, QtWidgets.QComboBox):
    """Self loading/saving QComboBox widget"""

    def __init__(self, db: dict, slices: Tuple, items: list):
        super(QtWidgets.QComboBox, self).__init__()
        AbstractDatabaseInterface.__init__(self, db, slices)
        self.addItems(items)
        self.load()
        self.currentTextChanged.connect(self._save)

    def load(self):
        self.setCurrentText(self._load())

    def value(self) -> str:
        return self.currentText()


class AutoFontComboBox(AbstractDatabaseInterface, QtWidgets.QFontComboBox):
    """Self loading/saving QFontComboBox widget"""

    def __init__(self, db: dict, slices: Tuple):
        super(QtWidgets.QFontComboBox, self).__init__()
        AbstractDatabaseInterface.__init__(self, db, slices)
        self.load()
        self.currentTextChanged.connect(self._save)

    def load(self):
        self.setCurrentText(self._load())

    def value(self) -> str:
        return self.currentText()


class AutoComboBoxGlob(AbstractDatabaseInterface, QtWidgets.QComboBox):
    """Self loading/saving QComboBox widget"""

    def __init__(self, db: dict, slices: Tuple, type_: str):
        super(QtWidgets.QComboBox, self).__init__()
        AbstractDatabaseInterface.__init__(self, db, slices)

        files = configEnum(type_)
        self.nameToStem = {str(f.name): f.stem.title() for f in files}
        self.stemToName = {f.stem.title(): str(f.name) for f in files}
        self.addItems(sorted([f.stem.title() for f in files]))
        self.load()
        self.currentTextChanged.connect(self._save)

    def load(self):
        try:
            stem = self.nameToStem[self._load()]
            self.setCurrentText(stem)
        except KeyError:
            log.error(f"'{self._load()}' not found")

    def value(self) -> str:
        name = self.stemToName[self.currentText()]
        return name


class AutoLineEdit(AbstractDatabaseInterface, QtWidgets.QLineEdit):
    """Self loading/saving QLineEdit widget"""

    def __init__(self, db: dict, slices: Tuple):
        super(QtWidgets.QLineEdit, self).__init__()
        AbstractDatabaseInterface.__init__(self, db, slices)
        self.load()
        self.textEdited.connect(self._save)

    def load(self):
        self.setText(self._load())

    def value(self) -> str:
        return self.text()


class AutoDoubleSpinBox(AbstractDatabaseInterface, QtWidgets.QDoubleSpinBox):
    """Self loading/saving QDoubleSpinBox widget"""

    def __init__(self, db: dict, slices: Tuple):
        super(QtWidgets.QDoubleSpinBox, self).__init__()
        AbstractDatabaseInterface.__init__(self, db, slices)
        self.setMaximum(1)
        self.setSingleStep(0.1)
        self.load()
        self.valueChanged.connect(self._save)

    def load(self) -> float:
        self.setValue(self._load())


class AutoSpinBox(AbstractDatabaseInterface, QtWidgets.QSpinBox):
    """Self loading/saving QSpinBox widget"""

    def __init__(self, db: dict, slices: Tuple):
        super(QtWidgets.QSpinBox, self).__init__()
        AbstractDatabaseInterface.__init__(self, db, slices)
        self.setMaximum(9999)
        self.load()
        self.valueChanged.connect(self._save)

    def load(self) -> int:
        self.setValue(self._load())


class Database(Observer):
    def __init__(self, path: Path, default: dict):
        self.timer = QtCore.QTimer(singleShot=True, interval=5000)
        self.timer.timeout.connect(self.save)
        super().__init__(self, callback=self.timer.start)
        self.path = path
        self.name = path.stem
        self.default = default
        self._setup()

    def save(self):
        """Saves dict content to JSON"""
        self.timer.stop()
        try:
            with open(self.path, "w", encoding="utf-8") as f:
                f.write(json.dumps(self, indent=2, sort_keys=False, cls=self.SetEncoder))
            log.info(f"Saved {self.name} database")
        except PermissionError:
            log.error(f"Could not save {self.name} database")

    def _load(self):
        """Loads JSON file content"""
        with open(self.path, encoding="utf-8") as f:
            self.update(json.load(f))
        log.info(f"Loaded {self.name} database")

    def _repair(self, db, default):
        """Repairs missing keys and type inconsistencies"""
        for key, value in default.items():
            if isinstance(value, dict):
                db.setdefault(key, value)
                if key not in REPAIR_EXCLUSIONS:
                    self._repair(db[key], value)
            else:
                db.setdefault(key, value)
            if not isinstance(db[key], type(value)):
                log.warning(f"Fixed type for '{key}': {type(db[key])} to {type(value)}")
                db[key] = default.get(key)

    def _setup(self):
        if self.path.is_file() and self.path.stat().st_size > 0:
            self._load()
            self._validate()
        else:
            self.path.parent.mkdir(parents=True, exist_ok=True)
            self.update(deepcopy(self.default))
            with open(self.path, "w", encoding="utf-8") as f:
                f.write(json.dumps(self, indent=2, sort_keys=False))
            log.info(f"Created {self.name} database '{self.path}'")

    def _validate(self):
        """Saves repairs to file (if any)"""
        old = json.dumps(self, sort_keys=True)
        default = deepcopy(self.default)
        self._repair(self, default)
        new = json.dumps(self, sort_keys=True)
        if not new == old:
            log.warning(f"Repaired missing keys in {self.name} database")
            self.save()

    class SetEncoder(json.JSONEncoder):
        def default(self, obj):
            """Converts sets and Paths"""
            if isinstance(obj, set):
                return list(obj)
            elif isinstance(obj, Path) or isinstance(obj, PurePath):
                return str(obj)
            return json.JSONEncoder.default(self, obj)



#!/usr/bin/python3
import re
from pathlib import Path
from typing import Callable
from PyQt5 import QtWidgets, QtGui
from PyQt5.QtGui import QTextCursor, QTextCharFormat
from PyQt5.QtWidgets import QTextEdit

try:
    from ..notes.text import AbstractText, AbstractTextBody, Polling
except (ValueError, ImportError):
    from notes.text import AbstractText, AbstractTextBody, Polling


class HTML(AbstractText, Polling):
    mode = "html"

    def __init__(self, core, path: Path):
        self.body = Body(self)
        super().__init__(core, path)
        self.body._init(self)

    @Polling.load  # Preload and decrypt file content as needed
    def load(self):
        """Loads file content"""
        if self.content.startswith("<!DOCTYPE HTML"):
            self.body.setHtml(self.content)
        elif not self.keyring.hasEncryption():
            html = self._plainToHtml()
            self.body.setHtml(html)
        else:
            self.body.setPlainText(self.content)

    def removeFontProperties(self, html: str) -> str:
        """Removes font-size and font-family style conflicts prior to save"""
        fontSize = r'(<body style="|<span style="|<p style=")(.*)(?:\sfont-size:\d+pt;)(.*">)'
        fontFamily = r'(<body style=")(.*)(?:\sfont-family:\'\w+\s*\w*\';)(.*">)'
        html = re.sub(fontSize, r"\1\2\3", html)
        html = re.sub(fontFamily, r"\1\2\3", html)
        return html

    def _plainToHtml(self) -> str:
        """Converts plain text to HTML format, preserving linebreaks and tabs"""
        text = self.content
        head = "<head><style>body { white-space: pre-wrap; }</style></head>"  # Preserve linebreaks height and tabs
        return f"<html>{head}<body>{text}</body></html>"


class Body(AbstractTextBody, QtWidgets.QTextEdit):
    def __init__(self, note):
        super().__init__()
        self.note = note
        self.setObjectName("html")
        self.setAcceptRichText(False)

    def insertFromMimeData(self, mime):
        """Accepts HTML content named 'qrichtext' (assumed to be an internal operation)"""
        accept = self.core.sdb["general"]["accept qrichtext"]
        isNative = re.findall(r'<\s*meta\s+name\s*=\s*"qrichtext".*?/>', mime.html())
        if accept and isNative and mime.hasHtml():
            self.insertHtml(mime.html())
        else:
            super().insertFromMimeData(mime)  # Insert as plain text

    def bold(self):
        Format.bold(self)

    def clearFormat(self):
        Format.clear(self)

    def highlight(self):
        format_ = Format(self)
        format_.highlight()

    def italic(self):
        Format.italic(self)

    def strike(self):
        Format.strike(self)

    def underline(self):
        Format.underline(self)


class Format:
    def __init__(self, body):
        self.qtext = body
        self.bg = body.note.css["highlight-bg"]
        self.fg = body.note.css["highlight-fg"]

    @classmethod
    def bold(cls, qtext):
        """Toggles text decoration for bold"""
        condition, operation = cls._hasBold, cls._setBold
        cls._toggleFormat(qtext, condition, operation)

    def highlight(self):
        """Toggles text decoration for highlighting"""
        condition, operation = self._hasHighlight, self._setHighlight
        self._toggleFormat(self.qtext, condition, operation)

    @classmethod
    def italic(cls, qtext):
        """Toggles text decoration for italic"""
        condition, operation = cls._hasItalic, cls._setItalic
        cls._toggleFormat(qtext, condition, operation)

    @classmethod
    def strike(cls, qtext):
        """Toggles text decoration for strikethrough"""
        condition, operation = cls._hasStrike, cls._setStrike
        cls._toggleFormat(qtext, condition, operation)

    @classmethod
    def underline(cls, qtext):
        """Toggles text decoration for underline"""
        condition, operation = cls._hasUnderline, cls._setUnderline
        cls._toggleFormat(qtext, condition, operation)

    @staticmethod
    def clear(qtext: QTextEdit):
        """Clears all text formatting"""
        charFormat = QTextCharFormat()
        charFormat.setFontItalic(False)
        charFormat.setFontStrikeOut(False)
        charFormat.setFontUnderline(False)
        charFormat.setFontWeight(QtGui.QFont.Normal)
        charFormat.setForeground(charFormat.foreground())
        charFormat.setBackground(charFormat.background())
        qtext.mergeCurrentCharFormat(charFormat)

    @staticmethod
    def _hasBold(charFormat: QTextCharFormat) -> bool:
        return charFormat.fontWeight() == QtGui.QFont.Bold

    @staticmethod
    def _hasHighlight(charFormat: QTextCharFormat) -> bool:
        return charFormat.background().isOpaque()

    @staticmethod
    def _hasItalic(charFormat: QTextCharFormat) -> bool:
        return charFormat.fontItalic()

    @staticmethod
    def _hasStrike(charFormat: QTextCharFormat) -> bool:
        return charFormat.fontStrikeOut()

    @staticmethod
    def _hasUnderline(charFormat: QTextCharFormat) -> bool:
        return charFormat.fontUnderline()

    @staticmethod
    def _setBold(charFormat: QTextCharFormat, enable: bool) -> QTextCharFormat:
        charFormat.setFontWeight(QtGui.QFont.Bold if enable else QtGui.QFont.Normal)
        return charFormat

    def _setHighlight(self, charFormat: QTextCharFormat, enable: bool) -> QTextCharFormat:
        default = QTextCharFormat()
        bg, fg = QtGui.QColor(self.bg), QtGui.QColor(self.fg)
        charFormat.setBackground(bg if enable else default.background())
        charFormat.setForeground(fg if enable else default.foreground())
        return charFormat

    @staticmethod
    def _setItalic(charFormat: QTextCharFormat, enable: bool) -> QTextCharFormat:
        charFormat.setFontItalic(enable)
        return charFormat

    @staticmethod
    def _setStrike(charFormat: QTextCharFormat, enable: bool) -> QTextCharFormat:
        charFormat.setFontStrikeOut(enable)
        return charFormat

    @staticmethod
    def _setUnderline(charFormat: QTextCharFormat, enable: bool) -> QTextCharFormat:
        charFormat.setFontUnderline(enable)
        return charFormat

    @classmethod
    def _toggleFormat(cls, qtext: QTextEdit, condition: Callable, operation: Callable):
        """Toggles a text decoration on the anchor or the selected text"""
        if qtext.textCursor().hasSelection():
            charFormat = cls._reverseSelectionFormat(qtext, condition, operation)
        else:  # Reverses anchor format
            enabled = condition(qtext.currentCharFormat())
            charFormat = QTextCharFormat()
            charFormat = operation(charFormat, not enabled)
        qtext.mergeCurrentCharFormat(charFormat)

    @staticmethod
    def _reverseSelectionFormat(
        qtext: QTextEdit, condition: Callable, operation: Callable
    ) -> QTextCharFormat:
        """Searches for a decoration in the selection then toggle its state"""
        charFormat = QTextCharFormat()
        cursor = qtext.textCursor()

        # Improves reliability compared to the assessment of the selection as a whole
        span = cursor.selectionEnd() - cursor.selectionStart()
        cursor.setPosition(cursor.selectionStart())
        for c in range(span):
            cursor.clearSelection()
            cursor.movePosition(QTextCursor.Right, QTextCursor.KeepAnchor)
            if condition(cursor.charFormat()):
                charFormat = operation(charFormat, False)
                break
        else:
            charFormat = operation(charFormat, True)
        return charFormat



#!/usr/bin/python3
import re
from collections import namedtuple
from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtWidgets import QLabel

try:
    from ..backend import logger
    from ..notes import NoteSetup
    from ..settings.base import Page, ActionBox, ActionTree
except (ValueError, ImportError):
    from backend import logger
    from notes import NoteSetup
    from settings.base import Page, ActionBox, ActionTree

Action = namedtuple("Action", ("key", "type", "icon"))
log = logger.new(__name__)


class PageHotkeys(Page):
    def __init__(self, *args):
        super().__init__(*args)
        self.addButton = QtWidgets.QPushButton("Add")
        self.addButton.clicked.connect(self._addNewItem)
        self.addButton.setEnabled(False)
        self.delButton = QtWidgets.QPushButton("Delete")
        self.delButton.clicked.connect(self._deleteSelectedItems)
        self.delButton.setEnabled(False)

        self.tree = Tree(self)
        self.tree.itemSelectionChanged.connect(self._treeSelectionChanged)
        self.actionBox = ActionBox(self.core)
        self.actionBox.populate()
        self.actionBox.currentTextChanged.connect(self._addEnabledUpdate)
        self.hotkeyEdit = KeySequenceEdit()
        self.hotkeyEdit.keyComboPressed.connect(self._addEnabledUpdate)
        self.hotkeyEdit.keyComboFailed.connect(self._addEnabledUpdate)

        layout = QtWidgets.QGridLayout(self)
        layout.addWidget(self.tree, 0, 0, 1, 4)
        layout.addWidget(QLabel("Action"), 1, 0, 1, 1)
        layout.addWidget(self.actionBox, 1, 1, 1, 3)
        layout.addWidget(QLabel("Hotkey"), 2, 0, 1, 1)
        layout.addWidget(self.hotkeyEdit, 2, 1, 1, 3)
        layout.addWidget(self.addButton, 3, 2, 1, 1)
        layout.addWidget(self.delButton, 3, 3, 1, 1)

    def _addEnabledUpdate(self):
        """Enables or disables add button"""
        seq = self.hotkeyEdit.sequence
        action = self.actionBox.currentText()
        for item in self.tree.iterate():
            if (item.text(0), item.text(1)) == (seq, action):
                self.addButton.setEnabled(False)
                return
        self.addButton.setEnabled(bool(seq))

    def _addNewItem(self):
        """Adds new hotkey item to tree widget and settings database"""
        seq = self.hotkeyEdit.sequence
        actionLabel = self.actionBox.currentText()
        if seq and actionLabel:
            self.tree.addItem(seq, actionLabel)
            self.tree.save()
            self._addEnabledUpdate()

    def _deleteSelectedItems(self):
        """Deletes selected items from tree and settings database"""
        self.tree.blockSignals(True)
        for i in self.tree.selectedItems():
            self.tree.deleteItem(i)
        self.tree.blockSignals(False)
        self.tree.save()

        self.delButton.setEnabled(False)
        self._addEnabledUpdate()

    def _treeSelectionChanged(self):
        """Updates combo box value from tree selection"""
        for i in self.tree.selectedItems():
            self.hotkeyEdit.sequence = i.text(0)
            self.actionBox.setCurrentText(i.text(1))
            break
        self.delButton.setEnabled(True)
        self._addEnabledUpdate()


class KeySequenceEdit(QtWidgets.QKeySequenceEdit):
    keyComboPressed = QtCore.pyqtSignal()
    keyComboFailed = QtCore.pyqtSignal()

    def __init__(self):
        super().__init__()
        self._reset()

    def keyPressEvent(self, event: QtGui.QKeyEvent):
        super().keyPressEvent(event)
        if self.sequence:
            seq = QtGui.QKeySequence(self.sequence)
            self.setKeySequence(seq)
            self.keyComboPressed.emit()
        else:
            self._reset()
            self.keyComboFailed.emit()

    @property
    def sequence(self) -> str:
        """Translates current key sequence into a plain hotkey string"""
        seq = self.keySequence().toString()
        seq = seq.split(",")[-1].strip()
        seq = "" if seq is None else seq
        if seq.count("+"):
            return seq

    @sequence.setter
    def sequence(self, text: str):
        """Sets current key sequence from a string"""
        seq = QtGui.QKeySequence.fromString(text)
        self.setKeySequence(seq)

    def _reset(self):
        """Clears current key sequence"""
        self.clear()
        lineEdit = self.findChild(QtWidgets.QLineEdit)
        lineEdit.setPlaceholderText("Press any key combination ...")


class Tree(ActionTree):
    def __init__(self, parent):
        super().__init__(parent)
        self.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self.setHeaderLabels(["Hotkey", "Action"])
        self._populate()

    def addItem(self, sequence: str, actionLabel: str):
        """Adds or updates an existing hotkey in the tree"""
        action = self.labelToAction(actionLabel)
        for item in self.iterate():  # Update hotkey in place if already present
            if item.text(0) == sequence:
                item.setText(1, actionLabel)
                item.setIcon(1, action.icon)
                self.setCurrentItem(item)
                break

        else:  # Add new item to tree
            index = self.currentIndex().row() + 1
            item = QtWidgets.QTreeWidgetItem()
            item.setText(0, sequence)
            item.setText(1, actionLabel)
            item.setIcon(1, action.icon)
            self.insertTopLevelItem(index, item)
            self.setCurrentItem(item)

    def deleteItem(self, item: QtWidgets.QTreeWidgetItem):
        """Removes the selected tree item"""
        index = self.indexOfTopLevelItem(item)
        self.takeTopLevelItem(index)

    def save(self):
        """Saves tree content to settings database"""
        keyEvents = {}
        for item in self.iterate():
            hotkey = self._formatHotkey(item.text(0))
            action = self.labelToAction(item.text(1))
            keyEvents[hotkey] = [action.type, action.key]
        self.db["key events"] = keyEvents

    def _formatHotkey(self, text: str) -> str:
        """Translates hotkeys from Mod+Mod+Key to mod,mod;key format"""
        key = re.sub(r"(\+)(?!.*\+)", ";", text.lower())  # Replace last occurence of '+'
        key = key.replace("+", ",")  # Replace all others occurences of '+'
        return key

    def _populate(self):
        """Parses database keys to generate QTreeWidget items"""
        noteActions = NoteSetup.actionsFromMode(self.core)
        coreActions = self.core.actions.tray
        for hotkey, command in self.db["key events"].items():
            try:
                type_, key = command
                action = noteActions[key] if type_ == "note" else coreActions[key]
                hotkey = re.sub("[,;]", "+", hotkey).title()
                item = QtWidgets.QTreeWidgetItem()
                item.setIcon(1, action.icon)
                item.setText(0, hotkey)
                item.setText(1, f"{action.label} ({type_})")
                self.addTopLevelItem(item)
            except KeyError:
                log.exception(f"Invalid action '{key}'")
        self.resizeColumnToContents(0)



#!/usr/bin/python3
import sys

try:
    from .__id__ import APP_NAME
    from .backend import UserDirs, PALETTE_DEFAULT, STYLE_DEFAULT
except (ValueError, ImportError):
    from __id__ import APP_NAME
    from backend import UserDirs, PALETTE_DEFAULT, STYLE_DEFAULT

warning = (
    f"This file is automatically generated, all changes will be lost. Close {APP_NAME} before editing."
)

CSS_DEFAULT = {
    "border-color": "#000000",
    "context-menu": "true",
    "encryption-icon": "#8e8e8e",
    "encryption-icon-active": "#8e8e8e",
    "highlight-bg": "#ffffa6",
    "highlight-fg": "#000000",
    "menu-icon": "#404040",
    "msgbar-visible": "true",
    "popup-visible": "true",
    "preview-bg": "#ffffff",
    "preview-fg": "#000000",
    "status-icon": "#bfbfbf",
    "status-icon-active": "#a8a8a8",
    "toolbar-align": "center",
    "toolbar-icon": "#bfbfbf",
    "toolbar-icon-active": "#bfbfbf",
    "toolbar-visible": "true",
    "tray-icon": "none",
}

NDB_DEFAULT = {"warning": warning, "favorites": [], "loaded": []}

SDB_DEFAULT = {
    "warning": warning,
    "general": {
        "default name": "Untitled",
        "skip taskbar": True,
        "hotkeys": True,
        "minimize": False,
        "reposition": True,
        "accept qrichtext": True,
        "encrypt all": False,
        "file monitor": False,
        "elide threshold": 60,
        "wheel threshold": 10,
        "truncate threshold": 35,
        "maximum load": 20,
        "ignored css": ["debug.css"],
    },
    "archives": {
        "frequency": 1,
        "image": True,
        "text": True,
    },
    "clean": {
        "archives": 90,
        "blanks": True,
        "logs": 7,
        "trash": 31,
    },
    "terminal": {
        "color scheme": "default.colorscheme",
        "font family": "DejaVu Sans Mono",
        "font size": 9,
    },
    "message bar": {
        "encryption": True,
        "folder": True,
        "pixels": True,
        "words": True,
        "popup interval": 1500,
    },
    "launcher": {
        "case": False,
        "hide": True,
    },
    "runtime": {
        "export directory": str(UserDirs.NOTES),
    },
    "profile default": {
        "position": [0, 0],
        "width": 300,
        "height": 210,
        "opacity": 1.0,
        "zoom": 0,
        "pin": False,
        "wrap": True,
        "style": STYLE_DEFAULT,
        "palette": PALETTE_DEFAULT,
        "mode": "html",
    },
    "key events": {
        "ctrl;a": ["note", "select all"],
        "ctrl;b": ["note", "bold"],
        "ctrl;c": ["note", "copy"],
        "ctrl;d": ["note", "line duplicate"],
        "ctrl;f": ["note", "search"],
        "ctrl;h": ["note", "highlight"],
        "ctrl;i": ["note", "italic"],
        "ctrl;m": ["note", "move"],
        "ctrl;p": ["note", "pin"],
        "ctrl;q": ["note", "unload"],
        "ctrl;r": ["note", "rename"],
        "ctrl;s": ["note", "strike"],
        "ctrl;u": ["note", "underline"],
        "ctrl;v": ["note", "paste"],
        "ctrl;w": ["note", "wrap"],
        "ctrl;x": ["note", "cut"],
        "ctrl;y": ["note", "redo"],
        "ctrl;z": ["note", "undo"],
        "ctrl;-": ["note", "zoom out"],
        "ctrl;=": ["note", "zoom in"],
        "ctrl,shift;backspace": ["note", "line delete"],
        "ctrl,shift;up": ["note", "line up"],
        "ctrl,shift;down": ["note", "line down"],
        "ctrl,shift;a": ["note", "select"],
        "ctrl,shift;c": ["note", "copy rich"],
        "ctrl,shift;l": ["note", "lowercase"],
        "ctrl,shift;p": ["note", "antidote corrector"],
        "ctrl,shift;r": ["note", "reset"],
        "ctrl,shift;s": ["note", "sort"],
        "ctrl,shift;t": ["note", "titlecase"],
        "ctrl,shift;u": ["note", "uppercase"],
        "ctrl,shift;v": ["note", "special paste"],
        "ctrl,shift;w": ["note", "clear format"],
        "ctrl,shift;x": ["note", "cut rich"],
    },
    "mouse events": {
        "tray": {
            "left": ["core", "toggle"],
            "middle": ["core", "new"],
        },
        "status": {
            "left": ["note", "swap"],
            "middle": ["core", "new"],
            "right": ["core", "menu"],
            "doubleclick": [],
            "up": ["note", "opacity+"],
            "down": ["note", "opacity-"],
        },
        "title": {
            "middle": ["note", "resize"],
            "right": [],
            "doubleclick": ["note", "rename"],
            "up": ["note", "roll"],
            "down": ["note", "unroll"],
        },
        "close": {
            "left": ["note", "unload"],
            "middle": [],
            "right": [],
            "doubleclick": [],
            "up": [],
            "down": [],
        },
    },
    "core menus": {
        "tray": [
            "new",
            "reset",
            "search",
            "open",
            "separator",
            "unload",
            "folders list",
            "separator",
            "notes list",
            "separator",
            "settings",
            "quit",
        ],
        "browser": [
            "separator",
            "new",
            "show",
            "load",
            "unload",
            "separator",
            "rename",
            "move",
            "open",
            "separator",
            "delete",
        ],
    },
    "context menus": {
        "html": [
            "new",
            "duplicate",
            "move",
            "export",
            "separator",
            "antidote corrector",
            "cut",
            "copy",
            "paste",
            "separator",
            "palette",
            "style",
            "encryption",
            "mode",
            "separator",
            "delete",
        ],
        "image": [
            "original",
            "scale",
            "aspect ratio",
            "fit",
            "separator",
            "duplicate",
            "move",
            "export",
            "separator",
            "palette",
            "style",
            "separator",
            "delete",
        ],
        "plain": [
            "new",
            "duplicate",
            "move",
            "export",
            "separator",
            "antidote corrector",
            "cut",
            "copy",
            "paste",
            "separator",
            "palette",
            "style",
            "encryption",
            "mode",
            "separator",
            "delete",
        ],
        "console": [
            "new",
            "duplicate",
            "move",
            "reload",
            "separator",
            "zoom in",
            "zoom out",
            "separator",
            "palette",
            "style",
            "mode",
            "separator",
            "delete",
        ],
        "vim": [
            "new",
            "duplicate",
            "move",
            "export",
            "separator",
            "zoom in",
            "zoom out",
            "separator",
            "palette",
            "style",
            "mode",
            "separator",
            "delete",
        ],
    },
    "toolbar menus": {
        "html": [
            "bold",
            "italic",
            "underline",
            "strike",
            "highlight",
            "clear format",
            "separator",
            "duplicate",
            "move",
            "lock|unlock",
            "separator",
            "delete",
        ],
        "image": [
            "original",
            "scale",
            "aspect ratio",
            "fit",
            "separator",
            "duplicate",
            "move",
            "palette",
            "style",
            "separator",
            "delete",
        ],
        "plain": [
            "new",
            "duplicate",
            "move",
            "lock|unlock",
            "separator",
            "antidote corrector",
            "palette",
            "style",
            "encryption",
            "mode",
            "separator",
            "delete",
        ],
        "console": [
            "new",
            "duplicate",
            "move",
            "reload",
            "separator",
            "zoom in",
            "zoom out",
            "separator",
            "palette",
            "style",
            "mode",
            "separator",
            "delete",
        ],
        "vim": [
            "new",
            "duplicate",
            "move",
            "separator",
            "zoom in",
            "zoom out",
            "separator",
            "palette",
            "style",
            "mode",
            "separator",
            "delete",
        ],
    },
    "legacy": 0.0,
}

if sys.platform.startswith("win"):
    SDB_DEFAULT["key events"]["ctrl;_"] = ["note", "zoom out"]



#!/usr/bin/python3
from pathlib import Path
from typing import Tuple
from PyQt5 import QtCore, QtWidgets, QtGui

try:
    from ..backend import logger
    from ..notes import Note
    from ..menus.bars import MessageLabel
except (ValueError, ImportError):
    from backend import logger
    from notes import Note
    from menus.bars import MessageLabel

log = logger.new(__name__)


class Image(Note):
    mode = "image"
    keepAspectRatio = False

    def __init__(self, core, path: Path):
        super().__init__(core, path)
        self._initNoteWindow(path)
        self.body = Body(self)
        self.gridLayout.addWidget(self.body, 1, 0, 1, 3)
        self.pixelsCounter = PixelsCounter(self)
        self.msgbarLayout.addWidget(self.pixelsCounter)

    def resizeEvent(self, event):
        """Allows to resize window regardless of image size"""
        self.body.setMinimumSize(1, 1)
        if self.body.isVisible():
            super().resizeEvent(event)

    def resizeFit(self):
        """Downsizes large image so they fit on screen"""
        screen = self.core.screen()
        iw, ih = self._size()
        sw, sh = screen.width(), screen.height()
        if iw * 1.2 > sw or ih * 1.2 > sh:
            factor = min(sw / iw, sh / ih) * 0.6
            width = int(iw * factor)
            height = int(ih * factor)
            self.core.pdb[self.id]["width"] = width
            self.core.pdb[self.id]["height"] = height
            self.resize(width, height)
            self.restoreAspectRatio()
        else:
            self._resizeOriginal()

    def resizeOriginal(self):
        """Restores original size"""
        self._resizeOriginal()
        self.restoreAspectRatio()  # Prevent imperfections

    def restoreAspectRatio(self):
        """Adjusts image size to fit original image aspect ratio"""
        # Redraw pixmap with proportional aspect ratio
        self.body.setScaledContents(False)
        self.body.redraw(QtCore.Qt.KeepAspectRatioByExpanding)

        # Adjust window to new scaled image size
        pixmap = self.body.pixmap()
        size = pixmap.width(), pixmap.height()
        self.body.setMinimumSize(*size)
        self.resize(1, 1)

        # Restore default resizing behavior
        self.body.setScaledContents(not self.keepAspectRatio)
        self.body.setMinimumSize(1, 1)

    def toClipboard(self):
        """Copies current image to system clipboard"""
        raise NotImplementedError  # child: _noteImageToClipboard

    def toggleAspectRatio(self):
        """Toggles 'keep aspect ratio' feature"""
        self.keepAspectRatio = not self.keepAspectRatio
        self.body.setScaledContents(not self.keepAspectRatio)
        self.body.redraw(QtCore.Qt.KeepAspectRatio)

        # Refresh toolbar latching icon
        self.setup.actions()
        self.setup.toolbar()

    def unsetSizeLimit(self):
        """Unsets minimum width and height"""
        self.setMaximumSize(QtWidgets.QWIDGETSIZE_MAX, QtWidgets.QWIDGETSIZE_MAX)
        self.setMinimumSize(0, 0)

    def _resizeOriginal(self):
        """Restores original size"""
        w, h = self._size()
        self.core.pdb[self.id]["width"] = w
        self.core.pdb[self.id]["height"] = h
        self.body.setMinimumSize(w, h)
        self.body.resize(w, h)

    def _savePNG(self, path: Path):
        """Saves curent image as a PNG file"""
        f = QtCore.QFile(self.path)
        f.open(QtCore.QIODevice.WriteOnly)
        image = QtGui.QPixmap(str(self.path))
        image.save(f, "PNG")

    def _size(self) -> Tuple[int, int]:
        # Returns original image size
        if not self.path.exists():
            self._savePNG(self.path)
        image = QtGui.QPixmap(str(self.path))
        return image.width(), image.height()


class Body(QtWidgets.QLabel):
    def __init__(self, note):
        super().__init__()
        self.note = note
        self.setObjectName("image")
        image = QtGui.QPixmap(str(note.path))
        policy = QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.MinimumExpanding

        self.setSizePolicy(*policy)
        self.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.setAlignment(QtCore.Qt.AlignCenter)
        self.setScaledContents(True)
        self.setMinimumSize(1, 1)
        self.setPixmap(image)

        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(note.menuTool)

    def focusInEvent(self, event):
        """Forwards focus event to Note"""
        self.note.focusInEvent(event)

    def focusOutEvent(self, event):
        """Forwards focus event to Note"""
        self.note.focusOutEvent(event)

    def redraw(self, aspect: QtCore.Qt.AspectRatioMode):
        """Redraws image at scale"""
        w, h = self.width(), self.height()
        transform = QtCore.Qt.SmoothTransformation
        image = QtGui.QPixmap(str(self.note.path))
        image = image.scaled(w, h, aspect, transformMode=transform)
        self.setPixmap(image)

    def resizeEvent(self, event):
        """Resizes with or without ratio correction"""
        if self.note.keepAspectRatio:
            self.redraw(QtCore.Qt.KeepAspectRatio)

    def zoomIn(self):
        """Increases note size"""
        self._zoom(1.05)

    def zoomOut(self):
        """Decreases note size"""
        self._zoom(0.95)

    def _zoom(self, factor: float):
        """Handles zoom in or out events"""
        w, h = self.note.width(), self.note.height()
        w, h = int(w * factor), int(h * factor)
        self.note.resize(w, h)
        self.note.saveGeometry()


class PixelsCounter(MessageLabel):
    def __init__(self, note):
        self.body = note.body
        super().__init__(note)
        self.setObjectName("msgbar-pixels")

    def _update(self):
        """Updates the pixel counter label for the current image"""
        w, h = self.body.width(), self.body.height()
        self.setText(f"{w} x {h} px")
        self.setVisible(self.core.sdb["message bar"]["pixels"])



#!/usr/bin/python3
import base64
import secrets
from cryptography.fernet import Fernet, InvalidToken
from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import QCryptographicHash
from PyQt5.QtWidgets import QLabel, QSizePolicy, QSpacerItem

try:
    from ..backend import logger, UserFiles
except (ValueError, ImportError):
    from backend import logger, UserFiles

log = logger.new(__name__)


class Cipher:
    MAGIC = "<ENCRYPTED/>"

    @staticmethod
    def key(password: str) -> Fernet:
        """Creates a Fernet key from a clear text password"""
        key = password.encode()
        for x in range(256**2):
            key = QCryptographicHash.hash((key), QCryptographicHash.Sha256)
        b64 = base64.urlsafe_b64encode(key)
        return Fernet(b64)

    @staticmethod
    def decrypt(key: bytes, token: bytes) -> str:
        """Decrypts a token using a Fernet key"""
        return key.decrypt(token).decode("utf-8")

    @staticmethod
    def encrypt(key: bytes, plain: str) -> bytes:
        """Encrypts text using a Fernet key"""
        return key.encrypt(plain.encode())


class Keyring:
    def __init__(self, note):
        self.note = note
        self.core = note.core

    def hasEncryption(self) -> bool:
        """Verifies if a file begins with an encryption tag"""
        with open(self.note.path, encoding="utf-8") as f:
            return f.read().startswith(Cipher.MAGIC)

    def hasKey(self) -> bool:
        """Returns true if text is decrypted or clear text"""
        try:
            text = self.note.body.toPlainText()
        except AttributeError:
            text = None
        text = text if text else self.note.content
        return not text.startswith(Cipher.MAGIC)

    def logout(self):
        """Resets entered creditals"""
        if not self.hasEncryption():
            return

        if self.hasKey():
            self.note.lockScreen.setEnabled(True)
            self.note.key = False  # False boolean to save the logged out state
            self.note.load()
            log.info(f"{self.note.id} : Logged out")

    def remove(self):
        """Removes encryption"""
        if self.hasKey():
            self.setKey(None)
            self.note.setup.msgbar()
            log.info(f"{self.note.id} : Removed encryption")
            self.note.popupFrame.showMessage("Removed encryption", "")

    def replace(self):
        """Replaces current key"""
        if not self.hasEncryption():
            self.setUnique()
            return
        elif not self.hasKey():
            return

        dialog = PasswordInputDialog(
            self.core, "Key replacement", f"Please provide a new password for\n{self.note.id}."
        )
        if dialog.exec_() == dialog.Accepted:
            self.setKey(dialog.key)
            log.info(f"{self.note.id} : Replaced unique key")
            self.note.popupFrame.showMessage("Replaced unique key", "")

    def setKey(self, key: Fernet):
        """Sets or removes a key, updates content and handles lock screen"""
        self.note.key = key
        self.note.save()
        self.note.load()
        self.note.setup.msgbar()
        try:
            if self.hasKey():
                self.note.lockScreen.setEnabled(False)
        except AttributeError:
            pass

    def setMaster(self):
        """Sets a master key"""
        # Current key is unknown
        if not self.hasKey():
            return

        # Master key is known
        elif self.core.masterKey:
            self.setKey(self.core.masterKey)
            log.info(f"{self.note.id} : Encrypted with master key")
            self.note.popupFrame.showMessage("Encrypted with master key", "")

        # Master key exist but is unknown
        elif UserFiles.KEYRING.is_file():
            if MasterKeyPrompt.login(self.core):
                self.setMaster()

        # Master key do not exist yet
        else:
            if MasterKeyPrompt.set(self.core):
                self.setMaster()

    def setUnique(self):
        """Sets a unique key"""
        if not self.hasKey():
            return

        dialog = PasswordInputDialog(
            self.core, "Encrypt with a unique key", f"Please provide a new password for\n{self.note.id}."
        )
        if dialog.exec_() == dialog.Accepted:
            self.setKey(dialog.key)
            log.info(f"{self.note.id} : Encrypted with unique key")
            self.note.popupFrame.showMessage("Encrypted with unique key", "")


class MasterKeyPrompt:
    @staticmethod
    def login(core) -> bool:
        """Shows a login prompt for master password"""
        dialog = MasterPasswordInputDialog(core, "Master Password", "Please provide the master password.")
        dialog.setWindowTitle("Primenote - Login")
        if dialog.exec_() == dialog.Accepted:
            core.masterKey = dialog.key
            log.info("Valid master key entered")
            return True
        else:
            log.warning("No master password provided")
            return False

    @staticmethod
    def save(core, key: Fernet):
        """Creates a file containing the new master key"""
        core.masterKey = key
        token = Cipher.encrypt(key, secrets.token_hex(16))
        with open(UserFiles.KEYRING, "w", encoding="utf-8") as f:
            f.write(token.decode("utf-8"))
        log.info(f"Master token created in '{UserFiles.KEYRING}'")

    @classmethod
    def set(cls, core) -> bool:
        """Shows a prompt to set a new master key"""
        dialog = PasswordInputDialog(
            core, "Set the master key", "Please provide a password for the master key."
        )
        if dialog.exec_() == dialog.Accepted:
            cls.save(core, dialog.key)
            return True
        return False


class PasswordLine(QtWidgets.QLineEdit):
    def __init__(self, core, focus: bool = True):
        super().__init__()
        self.focus = focus
        self.setObjectName("lockscreen-input")
        self.setEchoMode(self.Password)
        self.setTextMargins(5, 2, 5, 2)
        self.setAutoFillBackground(True)
        self.textChanged.connect(lambda: self.setErrorPalette(False))

        icon = core.icons["style"]
        passwordRevealAction = QtWidgets.QAction(icon, "Show password", self)
        self.addAction(passwordRevealAction, self.TrailingPosition)
        eyeButton = self.findChild(QtWidgets.QAbstractButton)
        eyeButton.pressed.connect(self._showPassword)
        eyeButton.released.connect(self._hidePassword)

    def showEvent(self, event):
        if self.focus:
            self.setFocus()
        super().showEvent(event)

    def setErrorPalette(self, enabled: bool):
        """Colorizes password input/confirmation fields"""
        enabled = "true" if enabled else "false"
        self.setProperty("error", enabled)
        self.style().polish(self)

    def _hidePassword(self):
        self.setEchoMode(self.Password)

    def _showPassword(self):
        self.setEchoMode(self.Normal)


class LockScreen(QtWidgets.QWidget):
    def __init__(self, note):
        super().__init__()
        self.setObjectName("lockscreen")
        self.note = note
        self.passwordLine = PasswordLine(note.core)
        self.passwordLine.returnPressed.connect(self._validate)

        frame = QtWidgets.QWidget()
        frame.setObjectName("lockscreen-frame")
        label = QLabel("The content of this file is encrypted. Please provide the key to continue.")
        label.setObjectName("lockscreen-label")
        label.setWordWrap(True)
        label.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        innerLayout = QtWidgets.QVBoxLayout(frame)
        innerLayout.addWidget(label)
        innerLayout.addWidget(self.passwordLine)

        layout = QtWidgets.QGridLayout(self)
        layout.addItem(QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding), 0, 1, 1, 1)
        layout.addItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum), 1, 0, 1, 1)
        layout.addWidget(frame, 1, 1, 1, 1)
        layout.addItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum), 1, 2, 1, 1)
        layout.addItem(QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding), 2, 1, 1, 1)
        self.setAutoFillBackground(True)
        self.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(note.menuTool)
        self.hide()

    def focusInEvent(self, event):
        """Forwards focus event to Note"""
        self.note.focusInEvent(event)
        super().focusInEvent(event)

    def focusOutEvent(self, event):
        """Forwards focus event to Note"""
        self.note.focusOutEvent(event)
        super().focusOutEvent(event)

    def paintEvent(self, event):
        """Provides stylesheet support for QWidget subclasses
        https://stackoverflow.com/a/18344643/4586648"""
        style = QtWidgets.QStyle.PE_Widget
        opt = QtWidgets.QStyleOption()
        opt.initFrom(self)
        painter = QtGui.QPainter(self)
        self.style().drawPrimitive(style, opt, painter, self)
        super().paintEvent(event)

    def setEnabled(self, enabled: bool):
        """Turns the lock screen on or off"""
        self.note.body.setVisible(not enabled)
        self.note.toolbarFrame.setVisible(not enabled)
        if bool(self.note.css["msgbar-visible"] == "true"):
            self.note.msgbarStack.setVisible(not enabled)
        self.note.showOuterFrames()
        self.setVisible(enabled)
        self.passwordLine.clear()

    def _validate(self):
        """Saves the key if its valid else throw an error"""
        key = Cipher.key(self.passwordLine.text())
        n = self.note
        try:
            with open(n.path, encoding="utf-8") as f:
                length = len(Cipher.MAGIC)
                token = f.read()[length:]
            Cipher.decrypt(key, token.encode())
            n.key = key
            n.load()
            self.setEnabled(False)
            if not n.rolled and n.hasClearText():
                n.setup.msgbar()
            log.info(f"{self.note.id} : Valid password provided")
        except InvalidToken:
            self.passwordLine.setErrorPalette(True)
            log.error(f"{self.note.id} : Invalid password")


class PasswordInputDialog(QtWidgets.QDialog):
    def __init__(self, core, title: str, label: str):
        super().__init__()
        self.passwordLine = PasswordLine(core)
        self.passwordLine.setPlaceholderText("Password")
        self.mirrorLine = PasswordLine(core, focus=False)
        self.mirrorLine.setPlaceholderText("Confirm password")

        frame = QtWidgets.QWidget()
        self.innerLayout = QtWidgets.QVBoxLayout(frame)
        self.innerLayout.addWidget(QLabel(label))
        self.innerLayout.addWidget(self.passwordLine)
        self.innerLayout.addWidget(self.mirrorLine)

        self.buttonBox = QtWidgets.QDialogButtonBox()
        self.buttonBox.setStandardButtons(self.buttonBox.Cancel | self.buttonBox.Save)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

        layout = QtWidgets.QGridLayout(self)
        layout.addItem(QSpacerItem(0, 100, QSizePolicy.Minimum, QSizePolicy.Expanding), 0, 1, 1, 1)
        layout.addItem(QSpacerItem(100, 0, QSizePolicy.Expanding, QSizePolicy.Minimum), 1, 0, 1, 1)
        layout.addWidget(frame, 1, 1, 1, 1)
        layout.addItem(QSpacerItem(100, 0, QSizePolicy.Expanding, QSizePolicy.Minimum), 1, 2, 1, 1)
        layout.addItem(QSpacerItem(0, 100, QSizePolicy.Minimum, QSizePolicy.Expanding), 2, 1, 1, 1)
        layout.addWidget(self.buttonBox, 3, 1, 1, 1)
        self.setWindowFlags(self.windowFlags() | QtCore.Qt.WindowStaysOnTopHint)
        self.setWindowTitle(title)
        self.resize(1, 1)

    def _validate(self) -> bool:
        """Saves the key if the password matches the confirmation field"""
        password = self.passwordLine.text()
        if not password:
            self.passwordLine.setErrorPalette(True)
        elif password != self.mirrorLine.text():
            self.mirrorLine.setErrorPalette(True)
        else:
            self.key = Cipher.key(password)
            return True
        return False

    def accept(self):
        if self._validate():
            super().accept()


class MasterPasswordInputDialog(PasswordInputDialog):
    def __init__(self, core, title: str, label: str):
        super().__init__(core, title, label)
        self.buttonBox.setStandardButtons(self.buttonBox.Cancel | self.buttonBox.Open)
        self.mirrorLine.hide()

    def accept(self):
        if self.passwordLine.text():
            self.key = Cipher.key(self.passwordLine.text())
            with open(UserFiles.KEYRING, encoding="utf-8") as f:
                token = f.read().encode()
            try:
                Cipher.decrypt(self.key, token)
                QtWidgets.QDialog.accept(self)
            except InvalidToken:
                self.passwordLine.setErrorPalette(True)
        else:
            self.passwordLine.setErrorPalette(True)



#!/usr/bin/python3
import random
import stat
from cryptography.fernet import InvalidToken
from pathlib import Path
from typing import Callable
from PyQt5 import QtCore
from PyQt5.QtGui import QTextCursor

try:
    from ..__id__ import APP_NAME
    from ..backend import Cursor, Tuples, logger
    from ..backend.cryptography import Cipher, LockScreen
    from ..notes import Note
    from ..menus.bars import WordsCounter
    from ..plugins import antidote
except (ValueError, ImportError):
    from __id__ import APP_NAME
    from backend import Cursor, Tuples, logger
    from backend.cryptography import Cipher, LockScreen
    from notes import Note
    from menus.bars import WordsCounter
    from plugins import antidote

log = logger.new(__name__)


class Polling:
    """Monitors file changes, loads and saves automatically"""

    def __init__(self):
        self.key = None
        self.saveTimer = QtCore.QTimer(interval=250, singleShot=True)
        self.saveTimer.timeout.connect(self.save)
        self.loadTimer = QtCore.QTimer(interval=5000)
        self.loadTimer.timeout.connect(self._loadSlot)
        self.loadTimer.start()

    def load(func: Callable) -> Callable:
        """Wrapper for the load method. Preloads content and update timestamp"""

        def wrapper(self, *args):
            if not self.path.is_file():
                return
            self._load()
            func(self, *args)
            self.saveTimer.stop()
            log.debug(f"{self.id} : File read")

        return wrapper

    def save(func: Callable) -> Callable:
        """Wrapper for the save method. Handles save timer, encryption and timestamp"""

        def wrapper(self, *args):
            self.saveTimer.stop()
            if not self.path.is_file():
                return
            self._save()
            func(self, *args)
            self._setTimestamp()
            log.debug(f"{self.id} : File saved")

        return wrapper

    def _decrypt(self) -> str:
        """Returns clear text file content"""
        keys = [k for k in (self.key, self.core.masterKey) if k]
        length = len(Cipher.MAGIC)
        token = self.content[length:]
        for key in keys:
            try:
                self.key = key
                return Cipher.decrypt(key, token.encode())
            except InvalidToken:
                pass
        return self.content

    def _encrypt(self) -> str:
        """Returns an encrypted version of the loaded content"""
        token = Cipher.encrypt(self.key, self.content)
        return Cipher.MAGIC + token.decode("utf-8")

    def _load(self):
        """Loads file content, handle encryption, updates timestamp"""
        with open(self.path, encoding="utf-8") as f:
            self.content = f.read()

        # Verifies if the note is encrypted and if key has been withdrawn (self.key == False)
        if self.keyring.hasEncryption() and self.key is not False:
            self.content = self._decrypt()
        self._setTimestamp()

    def _save(self):
        """Hook for note save method, handles encryption before saving"""
        if not self.body.toPlainText():
            self.content = ""
            return

        if self.hasClearText():
            if self.mode == "html":
                html = self.body.toHtml()
                html = self.removeFontProperties(html)
                self.content = html
            else:
                self.content = self.body.toPlainText()

            if self.key:
                self.content = self._encrypt()

    def _loadSlot(self):
        """Slot for the auto load timer"""
        self._setFrequency()
        if not self.isVisible():
            return
        if not self.path.is_file():
            return
        if self.mtime == self.path.stat().st_mtime:
            return
        log.warning(f"{self.id} : st_mtime mismatch ({self.mtime=} != {self.path.stat().st_mtime=})")
        self.load()

    def _setFrequency(self):
        """Sets polling frequency; minimizes IO for idle notes"""
        i = 250 if self.isVisible() else 5000
        self.loadTimer.setInterval(i)

    def _setTimestamp(self):
        """Updates last modification timestamp (inexpensive IO)"""
        try:
            self.mtime = self.path.stat().st_mtime
            log.debug(f"{self.id} : Updated timestamp")
        except FileNotFoundError:
            pass


class AbstractText(Note):
    def __init__(self, core, path: Path):
        super().__init__(core, path)
        self.core = core
        self._initNoteWindow(path)
        self._plugins()
        self.load()

        self.lockScreen = LockScreen(self)
        self.body.textChanged.connect(self.saveTimer.start)
        self.gridLayout.addWidget(self.body, 1, 0, 1, 3)
        self.gridLayout.addWidget(self.lockScreen, 1, 0, 1, 3)
        self.wordsCounter = WordsCounter(self)
        self.msgbarLayout.addWidget(self.wordsCounter)
        self.moveCursorToTop()

    def drop(self, mime: Tuples.Mime):
        """Handler for dropped text"""
        cursor = self.body.textCursor()
        cursor.insertText(mime.data)
        self.save()

    def focusInEvent(self, event):
        """Loads file content and forwards focus event to Note"""
        super().focusInEvent(event)
        if self.path.is_file():
            self._setWriteMode()

    def lock(self):
        """Sets file permission to read-only"""
        self._setReadOnly(True)

    @property
    def readOnly(self) -> bool:
        """Sets file write permissions for owner, group, others"""
        for mask in (stat.S_IWUSR, stat.S_IWGRP, stat.S_IWOTH):
            if self.path.stat().st_mode & mask:
                return False
        return True

    @Polling.save  # Handles save timer, encryption and timestamp
    def save(self):
        """Saves content to file"""
        try:
            with open(self.path, "w", encoding="utf-8") as f:
                f.write(self.content)
        except PermissionError:
            pass

    def search(self):
        self.core.search.attach(self.path)
        self.core.search.open()

    def showEvent(self, event):
        self.loadTimer.start(0)
        super().showEvent(event)
        if not self.hasClearText():
            self.lockScreen.setEnabled(True)

    def toggleLock(self):
        """Toggles read-only status"""
        self._setReadOnly(not self.readOnly)

    def unlock(self):
        """Sets file permission to read-write"""
        self._setReadOnly(False)

    def _plugins(self):
        """Loads third-party software"""
        try:
            self.antidote = antidote.getHandler(APP_NAME, self.body)
            self.antidote.init()
        except FileNotFoundError:
            log.debug("Could not find Antidote binary")

    def _setReadOnly(self, enabled: bool):
        """Sets file permission as read-only or read-write"""
        mask = stat.S_IWUSR | stat.S_IWGRP | stat.S_IWOTH
        mode = self.path.stat().st_mode
        mode = mode & ~mask if enabled else mode | mask
        try:
            self.path.chmod(mode)
        except PermissionError:
            pass
        self._setWriteMode()

    def _setWriteMode(self):
        """Updates the read-only / read-write status"""
        if self.readOnly != self.body.isReadOnly():
            self.body.setReadOnly(self.readOnly)
            self.setup.css()
            self.setup.actions()
            self.setup.toolbar()


class AbstractTextBody:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def dragEnterEvent(self, event):
        """Allows the implementation of a customized dropEvent handler"""
        if event.mimeData().hasFormat("text/plain"):
            event.acceptProposedAction()
        super().dragEnterEvent(event)

    def dropEvent(self, event):
        """Handler for dropped text"""
        super().dropEvent(event)
        self.note.save()
        self.note.activateWindow()

    def focusInEvent(self, event):
        """Forwards focus event to Note"""
        self.note.focusInEvent(event)
        super().focusInEvent(event)

    def focusOutEvent(self, event):
        """Forwards focus event to Note"""
        self.note.focusOutEvent(event)
        super().focusOutEvent(event)

    def keyPressEvent(self, event):
        """Hotkeys handler"""
        keyFilter = self.note.keyFilter
        keyFilter.update(event)
        if keyFilter.match():
            keyFilter.execute()
        elif not keyFilter.ignored():
            super().keyPressEvent(event)

    def antidoteCorrector(self):
        """Launches Antidote grammar corrector"""
        try:
            self.note.antidote.corrector()
        except AttributeError:
            log.error("Could not find Antidote binary")

    def antidoteDictionnary(self):
        """Launches Antidote dictionnaries"""
        try:
            self.note.antidote.dictionnary()
        except AttributeError:
            log.error("Could not find Antidote binary")

    def antidoteGuide(self):
        """Launches Antidote language guides"""
        try:
            self.note.antidote.guide()
        except AttributeError:
            log.error("Could not find Antidote binary")

    def capitalize(self):
        """Capitalize case of the selected text"""
        self._setCase(str.capitalize)

    def copyPlain(self):
        """Copies the selected plain text to system clipboard"""
        selected = self.textCursor().selectedText()
        selected = selected.replace("\0", "")  # Remove null bytes
        selected = selected.replace("\u2029", "\n")  # Replace Qt paragraph separators
        self.core.clipboard.setText(selected)

    def cutPlain(self):
        """Cuts the selected plain text to system clipboard"""
        self.copy()
        self.textCursor().removeSelectedText()

    def paste(self):
        pixmap = self.core.clipboard.image()
        if self.core.clipboard.mimeData().hasText():
            super().paste()
        elif not pixmap.isNull():
            log.info(f"{self.note.id} : Saving pasted pixmap ...")
            path = self.note.path
            if path.stat().st_size == 0:
                self.core.notes.delete(path)
            self.core.savePNG(pixmap)

    def pasteSpecial(self):
        """Pastes text from clipboard, replaces newlines with spaces"""
        text = self.core.clipboard.text()
        for linebreak in ("\r\n", "\n", "\t", "\r"):
            text = text.replace(linebreak, " ")
        self.insertPlainText(text)

    def lineDelete(self):
        """Deletes current block along with its newline char"""
        LineOperations(self).delete()

    def lineDuplicate(self):
        """Duplicates current block"""
        LineOperations(self).duplicate()

    def lineDown(self):
        """Shifts current block downward"""
        LineOperations(self).moveDown()

    def lineEnd(self):
        """Moves cursor to end of line"""
        LineOperations(self).toEnd()

    def lineUp(self):
        """Shifts current block upward"""
        LineOperations(self).moveUp()

    def lineSelect(self):
        """Select current block"""
        LineOperations(self).select()

    def lineSort(self):
        """Sorts or reverse-sorts current block"""
        LineOperations(self).sort()

    def lineShuffle(self):
        """Shuffles current block"""
        LineOperations(self).shuffle()

    def lineStart(self):
        """Moves cursor to start of line"""
        LineOperations(self).toStart()

    def lowercase(self):
        """Lowers the case of the selected text"""
        self._setCase(str.lower)

    def swapcase(self):
        """Swapcase the selected text"""
        self._setCase(str.swapcase)

    def titlecase(self):
        """Titlecase the selected text"""
        self._setCase(str.title)

    def uppercase(self):
        """Uppers the case of the selected text"""
        self._setCase(str.upper)

    def wrap(self, state=None):
        """Toggles or sets text wrap mode"""
        if state is None:
            self.wrap(not bool(self.lineWrapMode()))
        else:
            state = self.WidgetWidth if state else self.NoWrap
            self.setLineWrapMode(state)

    def zoomIn(self):
        """Increases font point size"""
        size = self.font().pointSize() + 1
        self.setStyleSheet(f"font-size: {size}pt")

    def zoomOut(self):
        """Decreases font point size"""
        size = max(5, self.font().pointSize() - 1)
        self.setStyleSheet(f"font-size: {size}pt")

    def _setCase(self, case: Callable):
        """Apply a string operation on the selection while preserving text decorations"""
        cursor = self.textCursor()
        cursor.beginEditBlock()
        old = Cursor(cursor.position(), cursor.anchor(), cursor.charFormat())  # Save selection

        # Loop over every chars of the selection
        text = case(cursor.selectedText())
        cursor.setPosition(cursor.selectionStart())
        for char in text:
            cursor.clearSelection()
            cursor.movePosition(QTextCursor.Right, QTextCursor.KeepAnchor)

            # Apply operation and restore format
            charFormat = cursor.charFormat()
            cursor.insertText(char)
            cursor.mergeCharFormat(charFormat)

        # Restore selection
        if old.anchor < old.pos:  # Anchor at left
            new = Cursor(cursor.position(), old.anchor)
        else:  # Anchor at right
            new = Cursor(old.pos, cursor.position())
        cursor.setPosition(new.anchor)
        cursor.setPosition(new.pos, QTextCursor.KeepAnchor)
        self.setTextCursor(cursor)
        cursor.endEditBlock()

    def _init(self, note):
        """Setups Q_TextEdit properties"""
        self.note = note
        self.core = note.core
        self.setAcceptDrops(True)
        self.setTabStopWidth(16)
        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(note.menuTool)


class LineOperations:
    def __init__(self, body):
        self.body = body
        self.core = body.core
        self.cursor = body.textCursor()
        self._selectBlock()

    def editBlock(func: Callable) -> Callable:
        """Groups editing operations as a single undo/redo change"""

        def wrapper(self, *args):
            self.cursor.beginEditBlock()
            func(self, *args)  # Execute line operation
            self.cursor.endEditBlock()
            self.body.setTextCursor(self.cursor)  # Apply changes

        return wrapper

    @editBlock
    def delete(self):
        """Deletes current block along with its newline char"""
        self.cursor.removeSelectedText()
        direction = QTextCursor.Left if self.cursor.atEnd() else QTextCursor.Right
        self.cursor.movePosition(direction, QTextCursor.KeepAnchor)
        self.cursor.removeSelectedText()

    @editBlock
    def duplicate(self):
        """Duplicates current block"""
        self.cursor.insertFragment(self.selection.fragment)
        self.cursor.insertText("\n")
        self.cursor.insertFragment(self.selection.fragment)
        self._select()

    @editBlock
    def moveDown(self):
        """Shifts current block downward"""
        if self.cursor.blockNumber() == self.body.document().blockCount() - 1:
            self._select()
        else:
            # Remove selection and take preceeding newline character (if any)
            self.cursor.removeSelectedText()
            self.cursor.movePosition(QTextCursor.Right, QTextCursor.KeepAnchor)
            self.cursor.removeSelectedText()

            # Insert saved text at the line below
            self.cursor.movePosition(QTextCursor.EndOfBlock)
            self.cursor.insertText("\n")
            self.cursor.insertFragment(self.selection.fragment)

            # Select inserted text
            length = len(self.selection.fragment.toPlainText())
            self.cursor.movePosition(QTextCursor.EndOfBlock)
            self.cursor.movePosition(QTextCursor.Left, QTextCursor.KeepAnchor, n=length)

    @editBlock
    def moveUp(self):
        """Shifts current block upward"""
        length = len(self.selection.fragment.toPlainText())
        if self.cursor.position() == length:
            self._select()
        else:
            # Remove selection and take preceeding newline character (if any)
            self.cursor.removeSelectedText()
            self.cursor.movePosition(QTextCursor.Left, QTextCursor.KeepAnchor)
            self.cursor.removeSelectedText()

            # Insert saved text at the line above
            self.cursor.movePosition(QTextCursor.StartOfBlock)
            self.cursor.insertFragment(self.selection.fragment)
            self.cursor.insertText("\n")

            # Select inserted text
            self.cursor.movePosition(QTextCursor.Left)
            self.cursor.movePosition(QTextCursor.Left, QTextCursor.KeepAnchor, n=length)

    def select(self):
        """Selects the current line"""
        pass  # Handled in __init__ by _selectBlock() method

    @editBlock
    def shuffle(self):
        """Shuffles current block"""
        self._reorder("shuffle")

    @editBlock
    def sort(self):
        """Sorts or reverse-sorts current block"""
        self._reorder("sort")

    def toEnd(self):
        """Moves cursor to end of line"""
        self.cursor.movePosition(QTextCursor.EndOfBlock)
        self.body.setTextCursor(self.cursor)

    def toStart(self):
        """Moves cursor to start of line"""
        self.cursor.movePosition(QTextCursor.StartOfBlock)
        self.body.setTextCursor(self.cursor)

    def _arrange(self, elements: list, mode: str) -> list:
        """Returns a list sorted with the choosen algorithm (sort/shuffle)"""
        # Sorted alphanumerical list, toggle reverse
        if mode == "sort":
            result = sorted(elements, key=self._sortKey)
            if result == elements:
                result = sorted(elements, key=self._sortKey, reverse=True)
            return result

        # Randomize
        elif mode == "shuffle":
            return random.sample(elements, len(elements))
        return []

    def _reorder(self, mode: str):
        """Sorts or shuffles current block"""
        text = self.selection.fragment.toPlainText()
        lines = text.splitlines()
        if len(lines) == 1:
            sep = self._separator(lines[0])
            parts = lines[0].split(sep) if sep else list(lines[0])
            parts = self._arrange(parts, mode)
            self.cursor.insertText(sep.join(parts))
        else:
            lines = self._arrange(lines, mode)
            for i, line in enumerate(lines):
                newline = "" if i == len(lines) - 1 else "\n"
                self.cursor.insertText(line + newline)
        self._select()

    def _select(self):
        """Highlights inserted text, place the anchor at block start"""
        self.cursor.setPosition(self.selection.pos)
        self.cursor.setPosition(self.selection.anchor, QTextCursor.KeepAnchor)

    def _selectBlock(self):
        """Extends and delimit the selection area. Prerequisite to self.selection"""
        cursor = self.cursor
        self.lineCount = cursor.selectedText().count("\u2029") + 1  # u2029: Qt paragraph separator
        if self.lineCount == 1:
            cursor.movePosition(QTextCursor.StartOfBlock)
            cursor.movePosition(QTextCursor.EndOfBlock, QTextCursor.KeepAnchor)
        else:
            selectionEnd = cursor.selectionEnd()
            cursor.setPosition(cursor.selectionStart())
            cursor.movePosition(QTextCursor.StartOfBlock)
            cursor.setPosition(selectionEnd, QTextCursor.KeepAnchor)
            cursor.movePosition(QTextCursor.EndOfBlock, QTextCursor.KeepAnchor)
        self.selection = Cursor(cursor.position(), cursor.anchor(), cursor.selection())
        self.body.setTextCursor(cursor)  # Apply the block selection

    def _separator(self, text: str) -> str:
        """Returns the separator with the highest occurence (if any)"""
        sep = {" ": 0, ",": 0, ";": 0, "|": 0}
        for s in sep:
            sep[s] = text.count(s)
        if sum(sep.values()) > 0:
            return max(sep, key=sep.get)
        return ""

    def _sortKey(self, key: str) -> tuple:
        """Sorts integers and strings from a list"""
        try:
            return (0, int(key))
        except ValueError:
            return (1, key.lower())



#!/usr/bin/python3
import os
import re
import sys
from collections import namedtuple
from dataclasses import dataclass
from glob import glob
from pathlib import Path
from PyQt5.QtGui import QTextFragment

try:
    from ..__id__ import ID
except (ValueError, ImportError):
    from __id__ import ID

DESKTOP_ENVIRONMENT = os.environ.get("DESKTOP_SESSION")
STYLE_DEFAULT = "minimal.css"
PALETTE_DEFAULT = "frozen.css"


class DeepPath(Path):
    _flavour = type(Path())._flavour

    def is_empty(self) -> bool:
        """Verifies if a file or a directory is empty"""
        if self.is_dir():
            empty = not [f for f in self.rglob("*")]
            return empty and not self.is_symlink()

        elif self.is_file():
            empty = self.stat().st_size == 0
            return empty

    def is_html(self) -> bool:
        """Verifies if the note file has HTML content"""
        with open(self, encoding="utf-8") as f:
            return f.read().startswith("<!DOCTYPE HTML")

    def is_nested(self) -> bool:
        """Verifies if the file is at the root of the note repository"""
        return self.relative_to(UserDirs.NOTES).parts[:-1]

    def is_note(self) -> bool:
        """Verifies if the path point to a note file"""
        return self.is_file() and self.suffix in (".png", ".txt")

    def is_relative_to(self, *other):  #
        """Returns True if the path is relative to another path or False
        Legacy support for Python 3.8.7 (added in Python 3.9)"""
        try:
            self.relative_to(*other)
            return True
        except ValueError:
            return False

    def is_script(self) -> bool:
        """Verifies if the note file is a script"""
        try:
            with open(self, encoding="utf-8") as f:
                return bool(re.findall(r"^#!(\/\w+.*)", f.read()))
        except UnicodeDecodeError:
            return False

    def rglob(self, pattern: str) -> list:
        """Search directory recursively"""
        # glob.glob temporarily replaces pathlib.glob to support directory links
        # see https://bugs.python.org/issue33428 (2018)
        path = str(self / "**" / pattern)
        found = glob(path, recursive=True)
        return [DeepPath(f) for f in found]


def configEnum(type_: str, ignored=[]) -> list:
    """Returns a filtered list of configuration files found in user and root folders"""
    Cfg = namedtuple("ConfigFolders", ("user", "root", "ext"))
    config = {
        "palette": Cfg(UserDirs.PALETTES, RootDirs.PALETTES, "css"),
        "style": Cfg(UserDirs.STYLES, RootDirs.STYLES, "css"),
        "terminal": Cfg(UserDirs.TERMINAL, RootDirs.TERMINAL, "colorscheme"),
    }[type_]

    user = [f for f in config.user.glob(f"*.{config.ext}") if f.is_file()]
    root = [f for f in config.root.glob(f"*.{config.ext}") if f.is_file()]
    rootNames = {f.name: f for f in root if f.name not in ignored}
    for f in user:
        if f.name in rootNames:
            del rootNames[f.name]
    root = [rootNames[n] for n in rootNames]
    return user + root


def userDir() -> Path:
    if sys.platform.startswith("win"):
        return DeepPath(Path.home() / ID)
    return DeepPath(Path.home() / ".config" / ID)


@dataclass
class UserDirs:
    CONFIG = userDir()
    NOTES = DeepPath(userDir() / "notes")
    ARCHIVES = CONFIG / "archives"
    LOGS = CONFIG / "logs"
    TRASH = CONFIG / "trash"
    UI = CONFIG / "ui"
    PALETTES = UI / "palettes"
    STYLES = UI / "styles"
    TERMINAL = UI / "terminal"
    VIM = UI / "vim"


@dataclass
class UserFiles:
    CSS = UserDirs.CONFIG / "ui" / "global.css"
    DECORATIONS = UserDirs.CONFIG / "ui" / "decorations.json"
    KEYRING = UserDirs.CONFIG / "master.key"
    LOG = UserDirs.LOGS / "session.log"
    NOTES = UserDirs.CONFIG / "notes.json"
    PROFILES = UserDirs.CONFIG / "profiles.json"
    SETTINGS = UserDirs.CONFIG / "settings.json"
    VIM = UserDirs.VIM / "vimrc"


@dataclass
class RootDirs:
    ROOT = Path(__file__).parents[1]
    FONTS = ROOT / "ui" / "fonts"
    ICONS = ROOT / "ui" / "icons"
    PALETTES = ROOT / "ui" / "palettes"
    STYLES = ROOT / "ui" / "styles"
    TERMINAL = ROOT / "ui" / "terminal"
    VIM = ROOT / "ui" / "vim"
    WIZARD = ROOT / "ui" / "wizard"


@dataclass
class RootFiles:
    CSS = RootDirs.ROOT / "ui" / "global.css"
    VIM = RootDirs.ROOT / "ui" / "vim" / "vimrc"
    DEFAULT_STYLE = RootDirs.STYLES / STYLE_DEFAULT
    DEFAULT_PALETTE = RootDirs.PALETTES / PALETTE_DEFAULT


@dataclass
class CoreActions:
    tray: dict
    browser: dict


@dataclass
class Cursor:
    pos: int
    anchor: int
    fragment: QTextFragment = None


@dataclass
class Tuples:
    Action = namedtuple("Action", ("label", "icon", "call"))
    Cleaner = namedtuple("Cleaner", ("dir", "delay"))
    Mime = namedtuple("Mime", ("type", "data"))
    Mode = namedtuple("Mode", ("icon", "label"))
    NoteIcons = namedtuple("NoteIcons", ("menu", "toolbar"))
    RegexCSS = namedtuple("RegexCSS", ("selector", "elements", "property"))
    SizeGrips = namedtuple("SizeGrips", ("left", "center", "right"))


def sanitizeFileName(text: str, default: str = ""):
    """Ensures that the filename are compliant with Windows and POSIX standards"""
    illegal = (
        "CON",
        "PRN",
        "AUX",
        "NUL",
        "COM1",
        "COM2",
        "COM3",
        "COM4",
        "COM5",
        "COM6",
        "COM7",
        "COM8",
        "COM9",
        "LPT1",
        "LPT2",
        "LPT3",
        "LPT4",
        "LPT5",
        "LPT6",
        "LPT7",
        "LPT8",
        "LPT9",
    )
    if text.upper() in illegal:
        return default

    for c in '\\/:*?"<>|':
        text = text.replace(c, "")
    text = text.replace("..", "")
    return text



#!/usr/bin/python3
import sys
from pathlib import Path
from PyQt5 import QtGui, QtWidgets
from PyQt5.QtCore import Qt

try:
    from ..settings.base import VSpacer
    from ..plugins import antidote
except (ValueError, ImportError):
    from settings.base import VSpacer
    from plugins import antidote


class PagePlugins(QtWidgets.QWidget):
    def __init__(self):
        super().__init__()
        Antidote = AntidoteWindows if sys.platform.startswith("win") else AntidoteLinux
        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(Terminal())
        layout.addWidget(Antidote())
        layout.addItem(VSpacer())


class Terminal(QtWidgets.QGroupBox):
    def __init__(self):
        super().__init__("QTermWidget")
        linuxBox = StatusCheckBox("Linux operating system")
        pyBox = StatusCheckBox("Python bindings")
        cppBox = StatusCheckBox("C++ backend")
        cppLabel = StatusLabel("/usr/lib/libqtermwidget5.so")
        cppLabel.setVisible(not sys.platform.startswith("win"))

        libPath = "/usr/lib/libqtermwidget5.so"
        cppBox.setChecked(Path(libPath).is_file())
        linuxBox.setChecked(not sys.platform.startswith("win"))
        pyBox.setChecked("QTermWidget" in sys.modules)

        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(linuxBox)
        layout.addWidget(cppBox)
        layout.addWidget(cppLabel)
        layout.addWidget(pyBox)


class AntidoteLinux(QtWidgets.QGroupBox):
    def __init__(self):
        super().__init__("Antidote grammar tool")
        binLabel_1 = StatusLabel("/usr/local/bin/AgentConnectix")
        binLabel_2 = StatusLabel("/usr/local/bin/Antidote")
        binBox = StatusCheckBox("C++ backend")

        isInstalled = Path(antidote.getBinaryPath()).is_file()
        binBox.setChecked(isInstalled)

        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(binBox)
        layout.addWidget(binLabel_1)
        layout.addWidget(binLabel_2)


class AntidoteWindows(QtWidgets.QGroupBox):
    def __init__(self):
        super().__init__("Antidote grammar tool")
        pyBox = StatusCheckBox("PyWin32 library")
        binBox = StatusCheckBox("C++ backend")
        binLabel = StatusLabel()

        pywin32 = True if "win32api" in sys.modules else False
        pyBox.setChecked(pywin32)

        antidotePath = antidote.getBinaryPath()
        binBox.setChecked(Path(antidotePath).is_file())
        binLabel.setVisible(bool(antidotePath))
        binLabel.setText(antidotePath)

        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(pyBox)
        layout.addWidget(binBox)
        layout.addWidget(binLabel)


class StatusLabel(QtWidgets.QLabel):
    def __init__(self, path: str = ""):
        super().__init__()
        font = QtGui.QFont()
        font.setItalic(True)
        font.setPointSize(7)
        self.setFont(font)
        self.setText(path)
        self.setIndent(20)
        self.setMouseTracking(True)
        self.setTextInteractionFlags(Qt.TextSelectableByMouse)


class StatusCheckBox(QtWidgets.QCheckBox):
    def __init__(self, text):
        super().__init__(text)
        self.setEnabled(False)
        self.setStyleSheet(
            """
            QCheckBox::indicator {
                width: 10px;
                height: 10px;
                border: 1px solid black;
            }
            QCheckBox:checked {
                color: darkgreen;
            }
            QCheckBox:unchecked {
                color: darkred;
            }
            QCheckBox::indicator:checked {
                    background-color: palegreen;
                }
                QCheckBox::indicator:unchecked {
                    background-color: darkred;
                }
                """
        )



#!/usr/bin/python3
import setuptools
import setuptools.command.build_py
import sys
from collections import namedtuple
from pathlib import Path

VERSION = "1.5"
DEPENDANCIES = ["pyqt5", "requests", "cryptography"]
PACKAGE_DATA = ["ui/*", "ui/*/*", "ui/*/*/*"]
APP_NAME = "PrimeNote"
ID = "primenote"
ALIAS = "pnote"

URL = f"https://gitlab.com/william.belanger/{ID}"
AUTHOR = "William Belanger"
DESCRIPTION = "Fanciest sticky note-taking application. Now with Vim!"
LONG_DESCRIPTION = (Path(__file__).parent / "README.md").read_text()
KEYWORDS = "sticky note plain rich text vim css pyqt qt"

PLATFORMS = ["Linux"]
CLASSIFIERS = [
    "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    "Programming Language :: Python :: 3.6",
    "Operating System :: POSIX :: Linux",
]

shortcuts = namedtuple("Shortcuts", ("title", "filename", "cmd", "icon"))
SHORTCUTS = [
    shortcuts(APP_NAME, f"{ID}.desktop", ALIAS, "primenote"),
    shortcuts(
        f"{APP_NAME} Settings", f"{ID}-settings.desktop", f"{ALIAS} -c settings", "primenote_settings"
    ),
]


class Shortcuts(setuptools.command.build_py.build_py):
    def run(self):
        for s in SHORTCUTS:
            path = Path(__file__).parent / s.filename
            with open(path, "w") as f:
                f.write(
                    "[Desktop Entry]\n"
                    f"Name={s.title}\n"
                    "Type=Application\n"
                    "Categories=Utility;\n"
                    f"Icon={s.icon}\n"
                    f"Exec={s.cmd}\n"
                )
        setuptools.command.build_py.build_py.run(self)


def datafiles():
    icons_path = [f"{ID}/ui/icons/{s.icon}.svg" for s in SHORTCUTS]
    links_path = [s.filename for s in SHORTCUTS]
    return [
        ("share/icons/hicolor/scalable/apps", icons_path),
        ("share/applications", links_path),
    ]


def setup():
    setuptools.setup(
        name=ID,
        version=VERSION,
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        long_description_content_type="text/markdown",
        keywords=KEYWORDS,
        author=AUTHOR,
        url=URL,
        classifiers=CLASSIFIERS,
        install_requires=DEPENDANCIES,
        platforms=PLATFORMS,
        cmdclass={"build_py": Shortcuts},
        data_files=datafiles(),
        package_data={"": PACKAGE_DATA},
        packages=setuptools.find_packages(),
        entry_points={"gui_scripts": [f"{ID}={ID}:main", f"{ALIAS}={ID}:main"]},
    )


if __name__ == "__main__":
    if sys.version_info < (3, 6):
        sys.exit("Python 3.6 or later is required to install this package")
    setup()



#!/usr/bin/python3
import json
import re
import requests
import shutil
import sys
import time
import zipfile
from argparse import Namespace
from cryptography.fernet import Fernet
from datetime import datetime, timedelta
from pathlib import Path
from typing import Iterable, List
from PyQt5 import QtWidgets, QtCore, QtGui

try:
    from .__id__ import APP_NAME, ID, VERSION
    from .__db__ import CSS_DEFAULT, SDB_DEFAULT, NDB_DEFAULT
    from .backend import (
        UserDirs,
        UserFiles,
        CoreActions,
        DeepPath,
        RootDirs,
        RootFiles,
        Tuples,
        cli,
        logger,
        sanitizeFileName,
    )
    from .backend.cli import CommandParser
    from .backend.cryptography import Cipher, MasterKeyPrompt
    from .backend.database import Database
    from .backend.observer import ObserverList
    from .backend.legacy import Legacy
    from .menus import MoveDialog
    from .menus.core import CoreMenu, ConfirmPrompt
    from .menus.search import Launcher, Search
    from .notes import Note
    from .notes.console import Console
    from .notes.vim import Vim
    from .notes.plain import Plain
    from .notes.html import HTML
    from .notes.image import Image
    from .settings import Settings
    from .settings.wizard import Wizard
except (ValueError, ImportError):
    from __id__ import APP_NAME, ID, VERSION
    from __db__ import CSS_DEFAULT, SDB_DEFAULT, NDB_DEFAULT
    from backend import (
        UserDirs,
        UserFiles,
        CoreActions,
        DeepPath,
        RootDirs,
        RootFiles,
        Tuples,
        cli,
        logger,
        sanitizeFileName,
    )
    from backend.cli import CommandParser
    from backend.cryptography import Cipher, MasterKeyPrompt
    from backend.database import Database
    from backend.observer import ObserverList
    from backend.legacy import Legacy
    from menus import MoveDialog
    from menus.core import CoreMenu, ConfirmPrompt
    from menus.search import Launcher, Search
    from notes import Note
    from notes.console import Console
    from notes.vim import Vim
    from notes.plain import Plain
    from notes.html import HTML
    from notes.image import Image
    from settings import Settings
    from settings.wizard import Wizard

HAS_TERMINAL = "QTermWidget" in sys.modules
log = logger.new(__name__)


class Core(QtCore.QObject):
    def __init__(self, app: QtWidgets.QApplication):
        super().__init__()
        self.app = app
        self.clipboard = app.clipboard()
        self.notes = NoteOperations(self)
        self.setup = CoreSetup(self)
        self.notes.init()
        app.primaryScreenChanged.connect(self.notes.reposition)
        log.info("Initialization completed")

    @property
    def loaded(self) -> dict:
        return self.notes.loaded

    def colorize(self, icon: QtGui.QIcon, color: QtGui.QColor) -> QtGui.QIcon:
        """Applies a custom foreground color on a monochrome QIcon"""
        pixmap = icon.pixmap(48, 48)
        painter = QtGui.QPainter(pixmap)
        painter.setCompositionMode(painter.CompositionMode_SourceIn)
        painter.fillRect(pixmap.rect(), color)
        painter.end()
        return QtGui.QIcon(pixmap)

    def execute(self, db: str, key: str, path: Path = None):
        """Translates mouse press events into core or note actions"""
        try:
            group, action = self.sdb["mouse events"][db][key]
            cmd = {group: [action]}
            if group == "note":
                cmd["note"].append(path)
            self.parser.fromDict(cmd)
        except (KeyError, ValueError):
            pass

    def fileManager(self, path: Path = UserDirs.NOTES):
        """Opens the default file system manager. Uses xdg-open for Linux or explorer for Windows"""
        path = path if path.is_dir() else path.parent
        self.fileManager.setArguments([str(path)])
        if not self.fileManager.startDetached()[0]:
            log.warning(f"Could not open directory '{path}'")

    def getNoteDecorationsCSS(self, *paths: Path) -> dict:
        """Serializes and combines the content of NoteDecorations{} selector from stylesheets"""
        css = dict(CSS_DEFAULT)
        if UserFiles.DECORATIONS.is_file():
            with open(UserFiles.DECORATIONS, encoding="utf-8") as f:
                css.update(json.load(f))

        stylesheets = [p for p in paths if p and p.is_file()]
        for s in stylesheets:
            with open(s, encoding="utf-8") as f:
                selector = re.findall(self.regex.selector, f.read())
                elements = re.findall(self.regex.elements, "".join(selector))
            for e in elements:
                css[e[0]] = e[1]
        return css

    def getNotesFiles(self, path: Path, recursive=True) -> list:
        """Returns a recursive list of note files. Accepts a directory"""
        files = []
        for ext in ("txt", "png"):
            if recursive:
                files += [f for f in path.rglob(f"*.{ext}") if f.is_file()]
            else:
                files += [f for f in path.glob(f"*.{ext}") if f.is_file()]
        return files

    def moveDialog(self, path: Path) -> Path:
        """Opens subfolder browser dialog"""
        dialog = MoveDialog(path)
        if dialog.exec_() == dialog.Accepted:
            dest = DeepPath(dialog.selectedFiles()[0]) / path.name
            if dest != path:
                dest = self.nameIndex(dest) if dest.exists() else dest
                self.notes.move(path, dest)

    def nameIndex(self, path: Path) -> Path:
        """Increments a numeric suffix until a unique name is found"""
        taken = [f.stem for f in path.parent.glob("*.txt") if f.exists()]
        taken += [f.stem for f in path.parent.glob("*.png") if f.exists()]
        taken += [f.name for f in path.parent.glob("*") if f.is_dir()]
        title = path.stem
        try:
            index = int(path.stem.rsplit(" ", 1)[1])
            title = path.stem.rsplit(" ", 1)[0]
        except (ValueError, IndexError):
            index = 1

        name = f"{title} {index}"
        while name in taken:
            name = f"{title} {index}"
            index += 1
        return (path.parent / name).with_suffix(path.suffix)

    def quit(self, *args):
        """Closes all notes and saves databases"""
        for path in dict(self.loaded):
            self.loaded[path].close()
        self.ndb.save()
        self.pdb.save()
        log.info(f"Leaving {APP_NAME}")
        self.app.quit()
        sys.exit(0)

    def savePNG(self, pixmap: QtGui.QPixmap):
        """Saves and loads a PNG file from a Mime object pixmap"""
        name = self.sdb["general"]["default name"] + ".png"
        path = self.nameIndex(UserDirs.NOTES / name)
        f = QtCore.QFile(str(path))
        f.open(QtCore.QIODevice.WriteOnly)
        pixmap.save(f, "PNG")
        self.notes.add(path)

    def searchToggle(self, show: bool = None):
        """Shows, hides or toggles the note search utility"""
        if show is None:
            visible = self.launcher.isVisible()
            self.launcher.setVisible(not visible)
        else:
            self.launcher.setVisible(show)
            self.launcher.activateWindow()

    def setGlobalIconsColor(self, key: str, color: str):
        """Sets a property override for NoteDecorations{} CSS"""

        def _reloadMenuIcons():
            self.setup.reloadIcons()
            for path, note in self.loaded.items():
                note.decorate()

        if UserFiles.DECORATIONS.is_file():
            with open(UserFiles.DECORATIONS, encoding="utf-8") as f:
                css = json.load(f)
        else:
            css = {}
        css[key] = color

        with open(UserFiles.DECORATIONS, "w", encoding="utf-8") as f:
            f.write(json.dumps(css, indent=2))

        if key == "tray-icon":
            self.tray.update()
        elif key == "menu-icon":
            _reloadMenuIcons()

    def settings(self):
        """Opens the settings dialog"""
        self.settingsDialog = Settings(self)

    def screen(self) -> QtGui.QScreen:
        """Returns the primary screen geometry"""
        screen = self.app.primaryScreen()
        return screen.availableGeometry()

    def screens(self) -> Iterable[QtGui.QScreen]:
        """Returns the all screens geometry"""
        for s in self.app.screens():
            yield s.availableGeometry()

    def trash(self, path: Path):
        """Moves a file or a folder tree to /trash/"""
        relative = path.relative_to(UserDirs.NOTES)
        dest = UserDirs.TRASH / relative
        dest.parent.mkdir(parents=True, exist_ok=True)
        if path.is_symlink():
            path.unlink()
        elif path.is_dir():
            shutil.copytree(path, dest, dirs_exist_ok=True)
            shutil.rmtree(path, ignore_errors=True)
        else:
            try:
                shutil.move(path, dest)
                log.info(f"Thrashed '{path}'")
            except FileNotFoundError:
                log.warning(f"trash() : No such file '{path}'")

    def wizard(self):
        """Launches wizard dialog"""
        self.wizardDialog = Wizard(self)
        self.wizardDialog.show()

    def _dropEvent(self, path: Path, mime: Tuples.Mime):
        """Handler for Mime objects parsed in a Note dropEvent. Accepts filesystem or web URL"""
        if path not in self.loaded:
            return  # Avoid race condition while saving dropped pixmap

        note = self.loaded[path]
        log.info(f"{note.id} : DropEvent : {mime}")

        if mime.type == "image":
            log.info(f"{note.id} : DropEvent : Saving dropped pixmap ...")
            if path.stat().st_size == 0:
                self.notes.delete(path)
            self.savePNG(mime.data)
        elif mime.type == "text":
            if note.mode in ("plain", "html", "vim"):
                note.drop(mime)

    def _popupMenu(self):
        """Opens the main contextual menu"""
        pos = QtGui.QCursor.pos()
        try:
            self.tray.menu.popup(pos)
        except AttributeError:
            log.error("System tray icon is not yet available")


class CoreSetup:
    """Setups instances and variables for Core class"""

    def __init__(self, core):
        self.core = core
        newUser = not UserFiles.SETTINGS.is_file()
        self.database()
        self.clean()
        self.regex()
        self.fonts()
        self.icons()
        self.actions()
        self.extra()
        self.encryption()
        if newUser:
            core.wizard()
        else:
            Legacy.menus(core)
            Legacy.css(core)
            Legacy.profiles(core)
            Legacy.renameSaveAs(core)
            Legacy.renameAntidote(core)
        core.sdb["legacy"] = VERSION

    def actions(self):
        """Translates action names into labels, icons and function calls"""
        c = self.core
        Action = Tuples.Action

        c.actions = CoreActions(
            {
                # Core actions
                "folders list": Action("Folders tree", c.icons["folder_active"], lambda: None),
                "hide": Action("Hide all", c.icons["hide"], c.notes.hideAll),
                "load": Action("Load folders", c.icons["load_folders"], c.notes.load),
                "menu": Action("Show main menu", c.icons["tray"], c._popupMenu),
                "new": Action("New note", c.icons["new"], c.notes.new),
                "notes list": Action("Orphan notes list", c.icons["plain_regular"], lambda: None),
                "open": Action("Open in file manager", c.icons["folder_open"], c.fileManager),
                "quit": Action("Quit", c.icons["quit"], c.quit),
                "reset": Action("Reset positions", c.icons["reset"], c.notes.reset),
                "reverse": Action("Reverse all", c.icons["reverse"], c.notes.toggleAll),
                "search": Action("Search repository", c.icons["search"], c.searchToggle),
                "separator": Action("Separator", c.icons["separator"], lambda: None),
                "settings": Action("Settings", c.icons["settings"], c.settings),
                "show": Action("Show all", c.icons["show"], c.notes.showAll),
                "toggle": Action("Toggle favorites", c.icons["toggle"], c.notes.toggleFavorites),
                "unload": Action("Unload folders", c.icons["unload_folders"], c.notes.unloadAllFolders),
                "wizard": Action("Setup wizard", c.icons["wizard"], c.wizard),
            },
            {
                # Browser actions
                "delete": Action("Delete folder", c.icons["delete"], c.notes.delete),
                "load": Action("Load all", c.icons["load_folders"], c.notes.load),
                "move": Action("Move folder", c.icons["move"], c.moveDialog),
                "new": Action("New note", c.icons["new"], c.notes.new),
                "open": Action("Open in file manager", c.icons["folder_open"], c.fileManager),
                "rename": Action("Rename folder", c.icons["rename"], c.notes.rename),
                "separator": Action("Separator", c.icons["separator"], lambda: None),
                "show": Action("Show all", c.icons["show"], c.notes.showAll),
                "unload": Action("Unload all", c.icons["unload_folders"], c.notes.unload),
            },
        )
        log.info("Loaded core actions")

    def clean(self):
        """Removes blanks, lint, expired files and manage archives"""
        self.records = self.RecordKeeping(self.core)
        self.recordsThread = QtCore.QThread()
        self.records.moveToThread(self.recordsThread)
        self.recordsThread.start()
        self.archiveTimer = QtCore.QTimer(interval=3600000)  # 60 min
        self.archiveTimer.timeout.connect(self._makeArchive)
        self.archiveTimer.start()
        self._makeArchive()

    def css(self):
        """Loads global.css stylesheet"""
        sheet = UserFiles.CSS if UserFiles.CSS.is_file() else RootFiles.CSS
        with open(sheet, encoding="utf-8") as f:
            self.core.app.setStyleSheet(f.read())
        log.debug(f"Loaded global CSS '{sheet}'")

    def database(self):
        """Initializes databases and favorites notes"""
        self.core.pdb = Database(UserFiles.PROFILES, default={})  # Profiles
        self.core.sdb = Database(UserFiles.SETTINGS, default=SDB_DEFAULT)  # Settings
        self.core.ndb = Database(UserFiles.NOTES, default=NDB_DEFAULT)  # Notes

    def encryption(self):
        """Prompts user for master key password"""
        self.core.masterKey = None
        if UserFiles.KEYRING.is_file():
            MasterKeyPrompt.login(self.core)

    def extra(self):
        """Setups auxiliaries instances"""
        self.core.parser = CommandParser(self.core)
        self.core.tray = TrayIcon(self.core)

        # Drop filter
        self.core.dropFilter = DropFilter()
        self.core.dropThread = QtCore.QThread()
        self.core.dropFilter.moveToThread(self.core.dropThread)
        self.core.dropFilter.mime.connect(self.core._dropEvent)
        self.core.dropThread.start()

        # File manager
        self.core.fileManager = self.FileManager()
        log.info("Loaded auxiliaries objects")

        # New files monitor
        self.core.fileMonitor = self.FileMonitorThread(self.core)
        self.core.fileMonitorThread = QtCore.QThread()
        self.core.fileMonitorThread.started.connect(self.core.fileMonitor.run)
        self.core.fileMonitor.moveToThread(self.core.fileMonitorThread)
        self.core.fileMonitor.newFile.connect(self.core.notes.add)
        self.core.fileMonitorThread.start()

        # Search file utilities
        self.core.search = Search(self.core)
        self.core.launcher = Launcher(self.core)

    def fonts(self):
        """Loads Liberation fonts for cross-platform stylesheets"""
        fonts = [str(f) for f in RootDirs.FONTS.glob("*.ttf") if f.is_file()]
        for f in fonts:
            QtGui.QFontDatabase.addApplicationFont(f)
        log.info("Loaded fonts")

    def icons(self):
        """Loads and colorize all SVG icons"""
        css = self.core.getNoteDecorationsCSS()
        color = QtGui.QColor(css["menu-icon"])
        self.core.icons = {}
        for path in [x for x in RootDirs.ICONS.glob("*.svg") if x.is_file()]:
            icon = QtGui.QIcon(str(path))
            self.core.icons[path.stem] = self.core.colorize(icon, color)
        log.info("Loaded icons")

    def regex(self):
        """Pre-compiles CSS regex patterns"""
        selector = re.compile(r"(?:NoteDecorations\s*\{\s*)([\s\S]*?)(?:\s*\})")  # NoteDecorations{}
        element = re.compile(r"([^:;\s]+)\s?:\s?([^;\s]+)(?=;)")  # Elements within a selector
        property_ = r"(NoteDecorations\s*\n*{[\w\W]*)%s\s*:\s*(#[abcdefABCDEF\d]{6}|\w+\s*);([\w\W]*?})"
        self.core.regex = Tuples.RegexCSS(selector, element, property_)

    def reloadIcons(self):
        """Reload icons in core and loaded notes"""
        self.icons()
        self.actions()
        for path in self.core.loaded:
            note = self.core.loaded[path]
            note.setup.actions()

    def _makeArchive(self):
        QtCore.QMetaObject.invokeMethod(self.records, "archive", QtCore.Qt.QueuedConnection)

    class FileManager(QtCore.QProcess):
        def __init__(self):
            super().__init__()
            self.setStandardOutputFile(QtCore.QProcess.nullDevice())
            self.setStandardErrorFile(QtCore.QProcess.nullDevice())
            if sys.platform.startswith("win"):
                self.setProgram("explorer")
            else:
                self.setProgram("xdg-open")

    class FileMonitorThread(QtCore.QObject):
        newFile = QtCore.pyqtSignal(Path)

        def __init__(self, core):
            super().__init__()
            self.core = core
            self.resume()

        def resume(self):
            self.current = set(self.core.getNotesFiles(UserDirs.NOTES))
            self.running = True

        def stop(self):
            self.running = False

        @QtCore.pyqtSlot()
        def run(self):
            while True:
                if self.running and self.core.sdb["general"]["file monitor"]:
                    files = set(self.core.getNotesFiles(UserDirs.NOTES))
                    for f in files - self.current:
                        self.newFile.emit(f)
                        log.info(f"Found externally created file '{f}'")
                    self.current = files
                time.sleep(1)

    class RecordKeeping(QtCore.QObject):
        def __init__(self, core):
            super().__init__()
            self.core = core
            self.blanks()
            self.purge()
            self.lint()

        @QtCore.pyqtSlot()
        def archive(self):
            """Creates a new archive if required"""
            if self.core.sdb["archives"]["frequency"]:
                UserDirs.ARCHIVES.mkdir(parents=True, exist_ok=True)
                limit = datetime.now() - timedelta(days=self.core.sdb["archives"]["frequency"])
                archives = [self._ctime(f) for f in UserDirs.ARCHIVES.glob("*.zip")]
                if not archives or max(archives) < limit:
                    self._saveArchive()

        def blanks(self):
            """Removes empty files and folders"""
            if self.core.sdb["clean"]["blanks"]:
                self._delete([f for f in UserDirs.NOTES.rglob("*") if f.is_empty()])

        def lint(self):
            """Removes dead keys from databases"""
            for nid in list(self.core.ndb["loaded"]):
                path = UserDirs.NOTES / nid
                if not path.exists():
                    self.core.ndb["loaded"].remove(nid)
                    log.warning(f"Removed '{nid}' from loaded database")

            for nid in list(self.core.ndb["favorites"]):
                path = UserDirs.NOTES / nid
                if not path.exists():
                    self.core.ndb["favorites"].remove(nid)
                    log.warning(f"Removed '{nid}' from favorites database")

            for nid in dict(self.core.pdb):
                path = UserDirs.NOTES / nid
                if not path.is_file():
                    del self.core.pdb[nid]
                    log.warning(f"Removed '{nid}' from profiles database")

            log.debug("Cleaned orphan databases entries")

        def purge(self):
            """Removes old files from trash, archives and logs folders"""
            tasks = (
                Tuples.Cleaner(UserDirs.TRASH, self.core.sdb["clean"]["trash"]),
                Tuples.Cleaner(UserDirs.LOGS, self.core.sdb["clean"]["logs"]),
                Tuples.Cleaner(UserDirs.ARCHIVES, self.core.sdb["clean"]["archives"]),
            )
            for task in tasks:
                expired = self._expired(task.dir, task.delay)
                self._delete(expired)
            log.info("Cleaned trashes, logs and archives")

        def _ctime(self, path: Path) -> datetime:
            """Returns the creation time of a file"""
            creation = path.stat().st_ctime
            return datetime.fromtimestamp(creation)

        def _delete(self, files: List[Path]):
            """Deletes a list of files or folders"""
            for f in files:
                try:
                    f.unlink()
                    log.info(f"Removed file '{f}'")
                except IsADirectoryError:
                    shutil.rmtree(f)
                    log.info(f"Removed folder '{f}'")
                except (PermissionError, FileNotFoundError):
                    pass

        def _expired(self, path: Path, days: int) -> List[Path]:
            """Recurses into a directory and returns a list of files older than <days>"""
            limit = datetime.now() - timedelta(days=days)
            try:
                return [f for f in path.rglob("*") if self._ctime(f) < limit]
            except FileNotFoundError as error:
                log.warning(f"{error}. Dead link ?")
            return []

        def _saveArchive(self):
            """Builds the archive"""
            files = []
            if self.core.sdb["archives"]["text"]:
                files += [f for f in UserDirs.NOTES.rglob("*.png") if f.is_file()]
            if self.core.sdb["archives"]["image"]:
                files += [f for f in UserDirs.NOTES.rglob("*.txt") if f.is_file()]

            if files:
                now = datetime.now()
                path = UserDirs.ARCHIVES / f"{now.year:02d}-{now.month:02d}-{now.day:02d}.zip"
                log.info(f"Saving new archive into '{path}' ...")
                self._zip(files, path)

        def _zip(self, files: List[Path], path: Path):
            """Creates a zip file while preserving the file tree"""
            try:
                with zipfile.ZipFile(path, "w", zipfile.ZIP_DEFLATED) as f:
                    for file in files:
                        f.write(file, file.relative_to(UserDirs.NOTES))
                log.info("Archive created successfully")
            except Exception:
                log.exception("Could not save archive")


class NoteOperations(QtCore.QObject):
    def __init__(self, core: QtCore.QObject):
        super().__init__()
        self.loaded = {}
        self.core = core

    @property
    def ndb(self) -> dict:
        return self.core.ndb

    @property
    def pdb(self) -> dict:
        return self.core.pdb

    @property
    def sdb(self) -> dict:
        return self.core.sdb

    @QtCore.pyqtSlot(Path)
    def add(self, path: Path, show=True, recursive=True):
        """Adds Note objects to self.loaded. Accepts a file or a folder tree"""
        if path.is_file():
            if path not in self.loaded:
                self._load(path)
                if show:
                    self.loaded[path].show()
                log.info(f"Loaded '{path}'")

        elif path.is_dir():
            for f in self.core.getNotesFiles(path, recursive=recursive):
                self.add(f)

    def addConfirm(self, path: Path):
        """Prevents the user before loading excessive amount of notes"""
        notesCount = len(self.core.getNotesFiles(path))
        if notesCount >= self.sdb["general"]["maximum load"]:
            msg = f"This action will load <b>{notesCount}</b> notes<br>all at once. Continue ?"
            dialog = ConfirmPrompt("Confirmation", msg)
            if dialog.exec_() == dialog.Yes:
                self.add(path)
        else:
            self.add(path)

    def delete(self, path: Path):
        """Sends note file to trash. Accepts a file or a folder tree"""
        self.unload(path)
        if path.exists():
            self.core.trash(path)
        self.core.setup.records.lint()

    def hideAll(self):
        """Hides all but pinned notes"""
        for path in dict(self.loaded):
            nid = self.loaded[path].id
            if not self.pdb[nid]["pin"]:
                self.loaded[path].hide()

    def init(self):
        """Loads subnotes and load/raise pinned notes on startup"""
        if not self.sdb["general"]["minimize"]:
            for nid, profile in self.pdb.items():
                if profile["pin"]:
                    self.add((UserDirs.NOTES / nid), show=True)

        for nid in list(self.ndb["loaded"] + self.ndb["favorites"]):
            self.add((UserDirs.NOTES / nid), show=False)

    def load(self, path=UserDirs.NOTES):
        """Loads notes recursively"""
        if bool(self.sdb["general"]["maximum load"]):
            self.addConfirm(path)
        else:
            self.add(path)

    def mode(self, path: Path, mode: str):
        """Sets a note interface mode (plain/html/vim/console)"""
        note = self.loaded[path]
        log.info(f"{note.id} : Changing mode from '{note.mode}' to '{mode}'")

        def htmlToPlainText(content: str, key: Fernet) -> str:
            textEdit = QtWidgets.QTextEdit()
            textEdit.setHtml(content)
            if key:
                token = Cipher.encrypt(key, textEdit.toPlainText())
                return Cipher.MAGIC + token.decode("utf-8")
            else:
                return textEdit.toPlainText()

        def setMode(note):
            isHtml = note.content.startswith("<!DOCTYPE HTML") or path.is_html()
            content = htmlToPlainText(note.content, note.key)
            note.save()

            self.pdb[note.id]["mode"] = mode
            note.close()
            if mode != "html" and isHtml:
                try:
                    with open(path, "w", encoding="utf-8") as f:
                        f.write(content)
                except PermissionError:
                    pass
            self.add(path)

        note.inModeTransition = True
        note.load()
        key = note.key
        setMode(note)  # Creates a new Note
        self.loaded[path].keyring.setKey(key)
        self.loaded[path].save()

    def move(self, src: Path, dest: Path):
        """Renames a note or a group of note. Accepts a file or a folder"""
        log.info(f"Moving '{src}' to '{dest}'")
        self.core.fileMonitor.stop()

        if src.is_file() and src.is_symlink():
            src.rename(dest)
            self._moveNoteMetadata(src, dest)

        elif src.is_file():
            self._moveNoteFile(src, dest)
            self._moveNoteMetadata(src, dest)

        elif src.is_dir() and src.is_symlink():
            for f in src.rglob("*"):
                self._moveNoteMetadata(f, dest / f.relative_to(src))
            src.rename(dest)

        elif src.is_dir():
            for f in src.rglob("*"):
                self.move(f, dest / f.relative_to(src))
            shutil.rmtree(src, ignore_errors=True)

        self.core.fileMonitor.resume()

    def new(self, path: Path = UserDirs.NOTES):
        """Create a new text note"""
        mode = self.sdb["profile default"]["mode"]
        name = self.sdb["general"]["default name"] + ".txt"
        dest = self.core.nameIndex(path / name)
        if not dest.parent.is_dir():  # Opened from a note action call
            dest = self.core.nameIndex(path.parent / name)  # Use note's parent path instead
        self.core.fileMonitor.stop()
        dest.touch()
        note = self._loadNote(dest, mode)
        note.show()

    def rename(self, path: Path):
        """Opens a renaming prompt. Accepts a file or a folder tree"""
        relative = path.relative_to(UserDirs.NOTES)
        dialog = QtWidgets.QInputDialog()
        dialog.setFixedSize(280, 120)
        dialog.setInputMode(dialog.TextInput)
        dialog.setWindowFlags(dialog.windowFlags() | QtCore.Qt.WindowStaysOnTopHint)
        dialog.setWindowTitle(f"Rename '{relative}'")
        dialog.setLabelText("Enter the new name:")
        dialog.setTextValue(path.stem)

        if dialog.exec_() == dialog.Accepted:
            dest = sanitizeFileName(dialog.textValue(), default=path.stem)
            dest = path.parent / (dest + path.suffix)
            if dest.exists() and dest != path:
                dest = self.core.nameIndex(dest)
            self.move(path, dest)

    def reposition(self, screen: QtGui.QScreen):
        """Moves out of bounds notes to the center of the primary screen"""
        if self.sdb["general"]["reposition"]:
            for path, note in self.loaded.items():
                if note.isVisible() and not note.onScreen:
                    note.snap()

    def reset(self):
        """Shows all and resets the geometry of all loaded notes"""

        def _move():
            total = len(self.loaded)
            for index, path in enumerate(self.loaded, start=1):
                note = self.loaded[path]
                pos = note.position(index, total)
                note.move(*pos)
                self.pdb[note.id]["position"] = pos

        def _opacity():
            for path, note in self.loaded.items():
                note.setWindowOpacity(1.0)
                self.pdb[note.id]["opacity"] = 1.0

        def _resize():
            for path, note in self.loaded.items():
                width = self.sdb["profile default"]["width"]
                height = self.sdb["profile default"]["height"]
                note.unroll()

                # Adjust image size prior to position calculation
                if note.mode == "image":
                    note.unsetSizeLimit()
                    note.resize(width, height)

                # Profile and stylesheet update
                self.pdb[note.id]["width"] = width
                self.pdb[note.id]["height"] = height
                note.resize(width, height)
                note.activateWindow()

        self.add(UserDirs.NOTES, recursive=False)
        QtCore.QTimer.singleShot(0, _opacity)
        QtCore.QTimer.singleShot(0, _resize)
        QtCore.QTimer.singleShot(0, _move)

    def showAll(self, folder: Path = UserDirs.NOTES):
        """Shows all loaded notes relative to a path"""
        for path, note in self.loaded.items():
            if folder in path.parents:
                note.show()

    def toggle(self, path: Path):
        """Show or hide a note"""
        if path not in self.loaded:
            self.add(path)
        else:
            isVisible = self.loaded[path].isVisible()
            self.loaded[path].setVisible(not isVisible)

    def toggleAll(self):
        """Toggles all loaded notes"""
        for path in self.loaded:
            self.toggle(path)

    def toggleFavorites(self):
        """Updates favorites notes list and toggle their visibility"""

        def _activatePinned():
            for path, note in self.loaded.items():
                if self.pdb[note.id]["pin"]:
                    note.activateWindow()

        def _visiblesSurvey() -> Iterable:
            for path, note in self.loaded.items():
                pinned = self.pdb[note.id]["pin"]
                if note.isVisible() and not pinned:
                    yield path

        def _relative(path: Path) -> str:
            relative = path.relative_to(UserDirs.NOTES)
            return str(relative)

        visibles = [_relative(path) for path in _visiblesSurvey()]
        favorites = [path for path in visibles if path in self.ndb["favorites"]]
        favorites = favorites if favorites else visibles
        if visibles:
            cb = self.ndb["favorites"].cb
            self.ndb["favorites"] = ObserverList(favorites, callback=cb)
            self.hideAll()
            _activatePinned()
        else:
            for path, note in self.loaded.items():
                if _relative(path) in self.ndb["favorites"]:
                    note.show()

    def unload(self, path: Path):
        """Cleanly closes Note objects. Accepts a file or a folder tree"""
        if path.is_dir():
            for f in self.core.getNotesFiles(path):
                self.unload(f)
        elif path in self.loaded:
            self.loaded[path].close()

    def unloadAllFolders(self):
        """Unloads all sub-folders"""
        folders = [f for f in UserDirs.NOTES.glob("*") if f.is_dir()]
        for f in folders:
            self.unload(f)

    def _load(self, path: Path):
        """Creates an image or a text Note instance from a file"""
        nid = str(path.relative_to(UserDirs.NOTES))
        try:  # Existing "mode" attribute in profile
            mode = self.pdb[nid]["mode"]
            note = self._loadNote(path, mode)
        except KeyError:  # New note
            if path.suffix == ".png":
                mode = "image"
            elif path.is_html():
                mode = "html"
            elif path.is_script() and HAS_TERMINAL:
                mode = "console"
            else:
                mode = self.sdb["profile default"]["mode"]  # Default

            log.info(f"New profile created for '{nid}' ({mode})")
            note = self._loadNote(path, mode)
            QtCore.QTimer.singleShot(0, note.resizeToContent)

        if path.is_nested():
            if nid not in self.ndb["loaded"]:
                self.ndb["loaded"].append(nid)

    def _loadNote(self, path: Path, mode: str) -> Note:
        """Creates an instance of the desired mode"""
        self.core.fileMonitor.stop()
        if mode == "image":
            self.loaded[path] = Image(self.core, path)
        elif mode == "html":
            self.loaded[path] = HTML(self.core, path)
        elif mode == "console" and HAS_TERMINAL:
            self.loaded[path] = Console(self.core, path)
        elif mode == "vim" and HAS_TERMINAL:
            self.loaded[path] = Vim(self.core, path)
        else:
            self.loaded[path] = Plain(self.core, path)
        self.core.fileMonitor.resume()

        note = self.loaded[path]
        if mode in ("html", "plain"):
            note.shown.connect(self.core.search.attach)
            note.hidden.connect(self.core.search.detach)
        return note

    def _moveNoteFile(self, src: Path, dest: Path):
        """Moves note in the file system"""
        dest.parent.mkdir(parents=True, exist_ok=True)
        try:
            shutil.move(src, dest)
        except (OSError, FileExistsError, PermissionError) as error:
            log.exception(error)

    def _moveNoteMetadata(self, src: Path, dest: Path):
        """Moves note metadata among databases. Called only AFTER the file was moved (Vim)"""
        oid = str(src.relative_to(UserDirs.NOTES))
        nid = str(dest.relative_to(UserDirs.NOTES))
        dest = DeepPath(dest)
        if src in self.loaded:  # Update loaded list
            self.loaded[dest] = self.loaded.pop(src)
            self.loaded[dest].setup.uid(dest)

            if oid in self.ndb["favorites"]:  # Update favorites
                self.ndb["favorites"].remove(oid)
                self.ndb["favorites"].append(nid)

            if isinstance(self.loaded[dest], Vim):  # Notice the Vim server
                self.loaded[dest].server.move(dest)

        if src in self.core.search.attached:  # Update attached list (search widget)
            self.core.search.attached.remove(src)
            self.core.search.attached.add(dest)

        if oid in self.pdb:  # Update the profiles database
            self.pdb[nid] = self.pdb.pop(oid)

        if dest in self.loaded:  # Update the message bar
            self.loaded[dest].folderLabel.update()
            self.loaded[dest].activateWindow()  # Restore window focus


class DropFilter(QtCore.QObject):
    mime = QtCore.pyqtSignal(object, object)

    def __init__(self):
        super().__init__()
        # This URL validator regular expression belongs to Django Software
        # Copyright (c) Django Software Foundation and individual contributors
        # Please read https://github.com/django/django/blob/master/LICENSE
        self.regex = re.compile(
            r"^(?:http|ftp)s?://"  # http:// or https://
            r"(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}(?<!-)\.?)|"  # domain
            r"localhost|"  # localhost
            r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}|"  # ipv4
            r"\[?[A-F0-9]*:[A-F0-9:]+\]?)"  # ipv6
            r"(?::\d+)?"  # optional port
            r"(?:/?|[/?]\S+)\Z",
            re.IGNORECASE,
        )

    @QtCore.pyqtSlot(object, str)
    def filter(self, path: Path, url: str) -> Tuples.Mime:
        """Returns the type & content of a filesystem or web URL"""
        url = self._sanitize(url)
        mime = Tuples.Mime("invalid", url)
        pixmap = QtGui.QPixmap(url)

        if self._isImage(pixmap):
            mime = Tuples.Mime("image", pixmap)

        elif self._isWebURL(url):
            req = self._request(url)
            pixmap.loadFromData(req)
            if self._isImage(pixmap):
                mime = Tuples.Mime("image", pixmap)
        else:
            try:
                with open(url, encoding="utf-8") as f:
                    mime = Tuples.Mime("text", f.read())
            except (UnicodeDecodeError, IsADirectoryError, PermissionError, OSError):
                pass
        self.mime.emit(path, mime)

    def _isImage(self, pixmap: QtGui.QPixmap) -> bool:
        return not pixmap.isNull()

    def _isWebURL(self, url: str) -> bool:
        return re.match(self.regex, url) is not None

    def _request(self, url: str) -> bytes:
        """Extracts content from a web URL"""
        try:
            return requests.get(url, timeout=5).content
        except requests.exceptions.RequestException:
            log.exception(f"Could not retrieve URL content '{url}'")

    def _sanitize(self, text: str) -> str:
        """Replaces hex codes from a string"""
        # ie. Replace '%5B~%5D' to '[~]'
        # %\d{2}       Patterns starting with a %, followed by two digit
        # %\d[a-fA-F]  Patterns starting with a %, followed by one digit and one letter from a to F
        hexChars = re.findall(r"%\d{2}|%\d[a-fA-F]", text)
        for code in hexChars:
            litteral = code.replace("%", "0x")
            litteral = chr(eval(litteral))
            text = text.replace(code, litteral)
        return text


class TrayIcon(QtWidgets.QSystemTrayIcon):
    def __init__(self, core):
        super().__init__()
        self.core = core
        self.activated.connect(self._clickEvent)
        self.update()

        # Fires a timer until the tray is available
        self.timer = QtCore.QTimer(singleShot=True, interval=1000)
        self.timer.timeout.connect(self._show)
        self.timer.start(0)

    def update(self):
        """Update tray icon svg and color"""
        css = self.core.getNoteDecorationsCSS()
        color = QtGui.QColor(css["tray-icon"])
        if color.isValid():
            # Overrides color if set in global.css
            icon = self.core.colorize(self.core.icons["tray"], color)
        else:
            path = str(RootDirs.ICONS / "tray.svg")
            icon = QtGui.QIcon(path)

        # QIcon to QPixmap conversion for KDE compatibility bug - FIXME: Still relevant ?
        pixmap = icon.pixmap(64, 64)
        self.setIcon(QtGui.QIcon(pixmap))

    def _clickEvent(self, button: QtWidgets.QSystemTrayIcon.ActivationReason):
        """Handler for mouse events on the tray icon"""
        key = {
            QtWidgets.QSystemTrayIcon.Trigger: "left",
            QtWidgets.QSystemTrayIcon.MiddleClick: "middle",
        }.get(button)
        self.core.execute("tray", key)

    def _show(self):
        """Waits until the system tray is available"""
        if self.isSystemTrayAvailable():
            self.menu = CoreMenu(self.core)
            self.setContextMenu(self.menu)
            self.show()
        else:
            self.timer.start()


def main(args: Namespace):
    """Initializes the application"""
    app = QtWidgets.QApplication(sys.argv)
    app.setDesktopFileName(ID)
    app.setApplicationName(APP_NAME)
    app.setQuitOnLastWindowClosed(False)
    core = Core(app)
    core.bus = cli.QDBusObject(core)
    core.parser.fromNamespace(args)
    app.exec()


if __name__ == "__main__":
    main()



#!/usr/bin/python3
from pathlib import Path
from PyQt5 import QtWidgets

try:
    from ..notes.text import AbstractText, AbstractTextBody, Polling
except (ValueError, ImportError):
    from notes.text import AbstractText, AbstractTextBody, Polling


class Plain(AbstractText, Polling):
    mode = "plain"

    def __init__(self, core, path: Path):
        self.body = Body()
        super().__init__(core, path)
        self.body._init(self)

    @Polling.load  # Preload and decrypt file content as needed
    def load(self):
        """Loads file content"""
        self.body.setPlainText(self.content)


class Body(AbstractTextBody, QtWidgets.QPlainTextEdit):
    def __init__(self):
        super().__init__()
        self.setObjectName("plain")
